import pytest

import datetime
import json
import requests

from eumdac.request import get, post, patch, put, delete


class MockResponse:
    def __init__(self, status_code, text=None):
        self.status_code = status_code
        self.text = text


def test_request_calls(monkeypatch):
    def mock_session_call(*args, **kwargs):
        return MockResponse(200)

    monkeypatch.setattr(requests.Session, "get", mock_session_call)
    monkeypatch.setattr(requests.Session, "post", mock_session_call)
    monkeypatch.setattr(requests.Session, "patch", mock_session_call)
    monkeypatch.setattr(requests.Session, "put", mock_session_call)
    monkeypatch.setattr(requests.Session, "delete", mock_session_call)

    assert get("https://test.test").status_code == 200
    assert post("https://test.test").status_code == 200
    assert patch("https://test.test").status_code == 200
    assert put("https://test.test").status_code == 200
    assert delete("https://test.test").status_code == 200


@pytest.mark.parametrize("message", [0, 1, 2])
def test_request_should_retry(message):
    jsons = [
        {
            "message": {
                "retryAfter": (datetime.timedelta(seconds=5) + datetime.datetime.now()).timestamp()
                * 1000,
                "reason": "test",
            }
        },
        {
            "message": {
                "reason": "Maximum number of connections exceeded",
            }
        },
        {
            "message": {
                "reason": "Maximum number of requests exceeded",
            }
        },
    ]
    from eumdac.request import _should_retry

    r = MockResponse(429, json.dumps(jsons[message]))
    assert _should_retry(r, 1)


@pytest.mark.parametrize("status_code", [200, 400, 500])
def test_request_should_not_retry(status_code):
    from eumdac.request import _should_retry

    r = MockResponse(status_code)
    assert not _should_retry(r)



"""Lockfile implementation mindful of OS specifics."""

import sys

from contextlib import contextmanager
from datetime import datetime, timedelta
from pathlib import Path
from threading import Lock
from time import sleep
from typing import Generator, IO, Optional

if sys.platform == "win32":
    import msvcrt
else:
    import fcntl


@contextmanager
def open_locked(
    lockfile_path: Path,
    timeout: Optional[timedelta] = None,
    delete: Optional[bool] = False,
) -> Generator[Optional[IO[str]], None, None]:
    """Open a file, locking it."""
    with open_locked.lock:  # type: ignore
        try:
            open_locked.locks  # type: ignore
        except AttributeError:
            open_locked.locks = {}  # type: ignore

        # create one lock object per file
        if lockfile_path not in open_locked.locks:  # type: ignore
            open_locked.locks[lockfile_path] = Lock()  # type: ignore
        lock = open_locked.locks[lockfile_path]  # type: ignore

    start = datetime.now()

    lockfile_path.parent.mkdir(exist_ok=True, parents=True)
    if timeout:
        r = lock.acquire(timeout=timeout.total_seconds())
    else:
        r = lock.acquire()
    if not r:
        yield None
    else:
        while True:
            if timeout and datetime.now() - start >= timeout:
                lockfile = None
                break

            lockfile = open(lockfile_path, "w")

            if sys.platform == "win32":
                try:
                    msvcrt.locking(lockfile.fileno(), msvcrt.LK_NBLCK, 1)
                    break
                except OSError:
                    sleep(0.1)
                    continue
            else:
                try:
                    fcntl.flock(lockfile, fcntl.LOCK_EX | fcntl.LOCK_NB)
                    break
                except IOError:
                    sleep(0.1)
                    continue

        yield lockfile

        if lockfile:
            if sys.platform == "win32":
                msvcrt.locking(lockfile.fileno(), msvcrt.LK_UNLCK, 1)
            else:
                fcntl.flock(lockfile, fcntl.LOCK_UN)
            lockfile.close()

            if delete:
                if sys.version_info >= (3, 8):
                    lockfile_path.unlink(missing_ok=False)
                else:
                    lockfile_path.unlink()
        lock.release()


setattr(open_locked, "lock", Lock())



"""Module containing helper functions related to eumdac configuration files"""

import os
from pathlib import Path


def get_config_dir() -> Path:
    """get the Path to the configuration directory of eumdac"""
    return Path(os.getenv("EUMDAC_CONFIG_DIR", (Path.home() / ".eumdac")))


def get_credentials_path() -> Path:
    """get the Path to the credentials of eumdac"""
    return get_config_dir() / "credentials"


def get_url_path() -> Path:
    """get the Path to the tailor configurations of eumdac"""
    r = get_config_dir() / "url"
    r.mkdir(parents=True, exist_ok=True)
    return r


PERCENTAGE_WARNING = 90



"""This module contains classes modeling Data Tailor resources."""

from __future__ import annotations

from collections.abc import Mapping
from dataclasses import asdict, dataclass
import requests
from typing import TYPE_CHECKING

from eumdac.request import _request
import eumdac.common

if TYPE_CHECKING:  # pragma: no cover
    import sys
    from typing import Any, Optional, Type, Union

    if sys.version_info < (3, 9):
        from typing import MutableMapping, Sequence
    else:
        from collections.abc import MutableMapping, Sequence
    from eumdac.datatailor import DataTailor

from eumdac.errors import EumdacError, eumdac_raise_for_status


def _none_filter(*args: Any, **kwargs: Any) -> MutableMapping[str, Any]:
    """Build a mapping of '*args' and '**kwargs' removing None values."""
    return {k: v for k, v in dict(*args, **kwargs).items() if v is not None}


class AsDictMixin:
    """Base class adding an 'asdict' method that removes None values."""

    def asdict(self) -> MutableMapping[str, Any]:
        """Return the fields of the instance as a new dictionary mapping field names to field values, removing None values."""
        return asdict(self, dict_factory=_none_filter)  # type: ignore


@dataclass
class Filter(AsDictMixin):
    """Layer filter, a list of `bands` or layers for a given `product`.

    Attributes
    ----------
    - `id`: *str*
    - `name`: *str*
        Human readable name.
    - `product`: *str*
        Product that the filter applies to.
    - `bands`: *list[dict]*
        List of bands part of the filter, as dicts of {id, number, name}.
    """

    __endpoint = "filters"
    id: Optional[str] = None
    bands: Optional[list] = None  # type: ignore[type-arg]
    name: Optional[str] = None
    product: Optional[str] = None


@dataclass
class RegionOfInterest(AsDictMixin):
    """Region of interest, a geographical area defined by its `NSWE` coordinates.

    Attributes
    ----------
    - `id`: *str*
    - `name`: *str*
        Human readable name.
    - `description`: *str*
        Human readable description.
    - `NSWE`:
        North, south, west, east coordinates, in decimal degrees.
    """

    __endpoint = "rois"
    id: Optional[str] = None
    name: Optional[str] = None
    NSWE: Optional[str] = None
    description: Optional[str] = None


@dataclass
class Quicklook(AsDictMixin):
    """Configuration for generating quicklooks."""

    __endpoint = "quicklooks"
    id: Optional[str] = None
    name: Optional[str] = None
    resample_method: Optional[str] = None
    stretch_method: Optional[str] = None
    product: Optional[str] = None
    format: Optional[str] = None
    nodatacolor: Optional[str] = None
    filter: Union[None, dict, Filter] = None  # type: ignore[type-arg]
    x_size: Optional[int] = None
    y_size: Optional[int] = None

    def __post_init__(self) -> None:
        """Prepare `filter` as a Filter instance if given as dict."""
        if self.filter is not None and isinstance(self.filter, dict):
            self.filter = Filter(**self.filter)


@dataclass
class Chain(AsDictMixin):
    """Chain configuration for Data Tailor customisation jobs."""

    __endpoint = "chains"
    __submodels = {"filter": Filter, "roi": RegionOfInterest, "quicklook": Quicklook}
    id: Optional[str] = None
    product: Optional[str] = None
    format: Optional[str] = None
    name: Optional[str] = None
    description: Optional[str] = None
    aggregation: Optional[str] = None
    projection: Optional[str] = None
    roi: Union[None, dict, RegionOfInterest] = None  # type: ignore[type-arg]
    filter: Union[None, dict, Filter] = None  # type: ignore[type-arg]
    quicklook: Union[None, dict, Quicklook] = None  # type: ignore[type-arg]
    resample_method: Optional[str] = None
    resample_resolution: Optional[list] = None  # type: ignore[type-arg]
    compression: Optional[dict] = None  # type: ignore[type-arg]
    xrit_segments: Optional[list] = None  # type: ignore[type-arg]

    def __post_init__(self) -> None:
        """Prepare attributes as an instance of their class if given as dict."""
        for name, Model in self.__submodels.items():
            attr = getattr(self, name)
            if attr is not None and isinstance(attr, Mapping):
                setattr(self, name, Model(**attr))


if TYPE_CHECKING:  # pragma: no cover
    CrudModelClass = Union[Type[Filter], Type[RegionOfInterest], Type[Quicklook], Type[Chain]]
    CrudModel = Union[Filter, RegionOfInterest, Quicklook, Chain]


class DataTailorCRUD:
    """Generic CRUD for Data Tailor models (Chain, ROI, Filter, Quicklook)."""

    datatailor: DataTailor
    Model: CrudModelClass
    endpoint: str
    url: str

    def __init__(self, datatailor: DataTailor, Model: CrudModelClass) -> None:
        """Init the CRUD for `datatailor` and `Model`."""
        self.datatailor = datatailor
        self.Model = Model
        endpoint = getattr(Model, f"_{Model.__name__}__endpoint")
        self.url = datatailor.urls.get("tailor", endpoint)

    def search(
        self, product: Optional[str] = None, format: Optional[str] = None
    ) -> Sequence[CrudModel]:
        """Search resources by 'format' and 'product'."""
        params = _none_filter(product=product, format=format)
        auth = self.datatailor.token.auth
        response = self._request("get", self.url, auth=auth, params=params)
        return [self.Model(**data) for data in response.json()["data"]]

    def create(self, model: CrudModel) -> None:
        """Create a new resource from 'model' on Data Tailor."""
        auth = self.datatailor.token.auth
        payload = model.asdict()
        self._request("post", self.url, auth=auth, json=payload)

    def read(self, model_id: str) -> CrudModel:
        """Retrieve the resource data with id 'model_id' from Data Tailor."""
        url = f"{self.url}/{model_id}"
        auth = self.datatailor.token.auth
        response = self._request("get", url, auth=auth)
        return self.Model(**response.json())

    def update(self, model: CrudModel) -> None:
        """Update the resource based on 'model' in Data Tailor."""
        data = model.asdict()
        url = f"{self.url}/{data['id']}"
        auth = self.datatailor.token.auth
        self._request("put", url, auth=auth, json=data)

    def delete(self, model: Union[str, CrudModel]) -> None:
        """Remove the resource 'model' from Data Tailor."""
        if isinstance(model, str):
            model_id = model
        else:
            model_id = model.id  # type: ignore[assignment]
        url = f"{self.url}/{model_id}"
        auth = self.datatailor.token.auth
        self._request("delete", url, auth=auth)

    def _request(self, method: str, url: str, **options: Any) -> requests.Response:
        """Perform a 'method' request to 'url' with 'options'."""
        response = _request(method, url, headers=eumdac.common.headers, **options)
        eumdac_raise_for_status(f"Request for {self.Model} failed.", response, EumdacError)
        return response



import pytest

from pathlib import Path
from dataclasses import asdict

from eumdac.tailor_models import Chain
from eumdac.order import (
    Order,
    new_order_filename,
    latest_order_file,
    all_order_filenames,
    highest_number_in_order_filenames,
    resolve_order,
    get_default_order_dir,
)

from tests.base import FakeProduct, FakeCollection, FakeCustomisation, FakeTailor

order_test_path = Path("test_order.yml")

fake_chain = Chain()
fake_chain.id = "id"
fake_chain.description = "description"


@pytest.fixture
def products():
    return [FakeProduct()]


@pytest.fixture
def tailor_order(tmp_path, products):
    order = Order(order_file=tmp_path / "test_tailor_order.yml")
    order.initialize(
        fake_chain,
        products=products,
        output_dir=tmp_path / "output",
        file_pattern=None,
        query=None,
    )
    yield order

    order.delete()


@pytest.fixture
def download_order(tmp_path, products):
    order = Order(order_file=tmp_path / "test_download_order.yml")
    order.initialize(
        None, products=products, output_dir=tmp_path / "output", file_pattern=None, query=None
    )
    yield order

    order.delete()


def test_order_status(tailor_order, products):
    assert tailor_order.status() == "NOT COMPLETED"

    for p in products:
        tailor_order.update(None, p._id, "FAILED", None)
    assert tailor_order.status() == "FAILED"

    for p in products:
        tailor_order.update(None, p._id, "DONE", None)
    assert tailor_order.status() == "NOT COMPLETED"

    for p in products:
        tailor_order.update(None, p._id, "DONE", {"fn": "DOWNLOADED"})
    assert tailor_order.status() == "DONE"


def test_order_collections(tailor_order, products):
    for p in products:
        assert p.collection._id in tailor_order.collections()


def test_order_pretty(tailor_order, download_order):
    for order in [tailor_order, download_order]:
        ps = order.pretty_string(True)
        assert str(order) in ps
        assert order.status() in ps
        assert str(order.collections()) in ps


def test_order_deserialization(tailor_order):
    with tailor_order.dict_from_file() as order_dict:
        assert isinstance(order_dict, dict)
        assert "output_dir" in order_dict
        assert "chain" in order_dict
        assert "products_to_process" in order_dict
        for ptp in order_dict["products_to_process"].values():
            assert ptp["col_id"] is not None
            assert ptp["server_state"] == "UNSUBMITTED"
            assert ptp["customisation"] is None


def test_order_update(tmp_path):
    p = tmp_path / order_test_path

    collection_id = "test_col_id"
    product_id = "test_prod_id"
    status = "test_status"
    download_states = {"test": "test_download_state"}
    query = {"test": "test_query_parameter"}
    products = [FakeProduct(id=product_id, collection=FakeCollection(id=collection_id))]
    order = Order(order_file=p)
    order.initialize(
        fake_chain,
        products=products,
        output_dir=tmp_path / "output",
        file_pattern=None,
        query=query,
    )
    order.update(collection_id, product_id, status, download_states)

    with order.dict_from_file() as order_dict:
        assert isinstance(order_dict, dict)
        assert order_dict["query"] == query
        for ptp in order_dict["products_to_process"].values():
            assert ptp["server_state"] == status
            assert ptp["download_states"] == download_states

    if p.exists():
        p.unlink()


def test_order_remote_delete_failed(tailor_order, products):
    with tailor_order.dict_from_file() as order_dict:
        for ptp in order_dict["products_to_process"].values():
            ptp["server_state"] = "FAILED"
    tailor_order._locked_serialize(order_dict)

    for p in products:
        tailor_order.update("test_customisation", p._id, "FAILED", {"fn": "DOWNLOADED"})

    tailor_order.remote_delete_failed(FakeTailor())


def test_order_resolve_product_num(tailor_order, products):
    cnt = 1
    for p in products:
        total_products, num = tailor_order.resolve_product_num(product_id=p._id)
        assert total_products >= num
        assert num == cnt
        cnt += 1

    with pytest.raises(KeyError):
        _, __ = tailor_order.resolve_product_num(product_id="INVALID")


def test_order_filehandling(tmp_path):
    path_order_dir = get_default_order_dir()
    assert path_order_dir.exists()

    order_filename1 = new_order_filename(path_order_dir)
    assert order_filename1.exists()

    order = Order(order_filename1)
    collection_id = "test_col_id"
    product_id = "test_prod_id"
    products = [FakeProduct(id=product_id, collection=FakeCollection(id=collection_id))]
    order.initialize(
        fake_chain,
        products=products,
        output_dir=tmp_path / "output",
        file_pattern=None,
        query=None,
    )

    assert order_filename1.exists()

    order_filename2 = new_order_filename(path_order_dir)
    assert order_filename2.exists()
    assert order_filename2 != order_filename1

    order = Order(order_filename2)
    collection_id = "test_col_id"
    product_id = "test_prod_id"
    products = [FakeProduct(id=product_id, collection=FakeCollection(id=collection_id))]
    order.initialize(
        fake_chain,
        products=products,
        output_dir=tmp_path / "output",
        file_pattern=None,
        query=None,
    )

    assert order_filename2.exists()

    o = resolve_order(path_order_dir, "latest")
    assert o._order_file == order._order_file

    if order_filename1.exists():
        order_filename1.unlink()
    if order_filename2.exists():
        order_filename2.unlink()



"""Module containing classes to handle the token authentication."""

from __future__ import annotations

import abc
import sys
import time
from configparser import ConfigParser
from datetime import datetime
from typing import TYPE_CHECKING, NamedTuple
from urllib.parse import quote as url_quote

import requests
from importlib import resources as importlib_resources
from requests.auth import AuthBase, HTTPBasicAuth

from eumdac.request import post
import eumdac.common
from eumdac.errors import EumdacError

from eumdac.logging import logger

if TYPE_CHECKING:  # pragma: no cover
    from typing import Optional

    if sys.version_info < (3, 9):
        from typing import Iterable, Mapping
    else:
        from collections.abc import Iterable, Mapping


class URLs(ConfigParser):
    """Dictionary-like ConfigParser based storage of EUMDAC related URLs"""

    def __init__(self, inifile: Optional[str] = None) -> None:
        super().__init__()
        if inifile:
            self.read(inifile)
        else:
            if sys.version_info >= (3, 9):
                with importlib_resources.as_file(
                    importlib_resources.files("eumdac") / "endpoints.ini"
                ) as path:
                    self.read(path)
            else:  # python < 3.9
                with importlib_resources.path("eumdac", "endpoints.ini") as path:
                    self.read(path)

    def get(  # type: ignore[override]
        self,
        section: str,
        option: str,
        raw: bool = False,
        vars: Optional[Mapping[str, str]] = None,
        fallback: str = "",
    ) -> str:
        """Get an option value for the given section"""
        if vars is not None:
            vars = {k: url_quote(str(v).encode()).replace("%", "%%") for k, v in vars.items()}
        return super().get(section, option, raw=raw, vars=vars, fallback=fallback)


class Credentials(NamedTuple):
    """Pair of Consumer Key and Secret authentication parameters.

    Attributes
    ----------
    - `consumer_key` : *str*
    - `consumer_secret` : *str*
    """

    consumer_key: str
    consumer_secret: str


class HTTPBearerAuth(AuthBase):
    """Attaches HTTP Bearer Authentication to the given Request object.

    Attributes
    ----------
    - `token`: *str*
        Bearer token
    """

    def __init__(self, token: str) -> None:
        """
        Parameters
        ----------
        - `token` : *str*
            Token to use for authentication
        """
        self.token = token

    def __call__(self, request: requests.PreparedRequest) -> requests.PreparedRequest:
        """Returns the given 'request' with the Bearer authentication parameter attached to the headers."""
        request.headers["authorization"] = f"Bearer {self.token}"
        return request


class BaseToken(metaclass=abc.ABCMeta):
    """Base class from which all eumdac authentication implementations derive"""

    urls: URLs

    @property
    def auth(self) -> Optional[AuthBase]:
        """To be overloaded in subclasses, shall return a configured AuthBase instance."""
        # overload in subclasses
        pass


class AccessToken(BaseToken):
    """EUMETSAT API access token

    Handles requesting of API tokens and their renewal after expiration.
    The str representation of AccessToken instances will be the current token value.

    Attributes
    ----------
    - `request_margin` : *int*
        seconds before expiration to start requesting a new token
    """

    request_margin: int = 2  # seconds
    _expiration: int = 0
    _access_token: str = ""

    credentials: Credentials
    validity_period: int  # seconds
    urls: URLs
    cache: bool  # does nothing, cache is always on in the server

    def __init__(
        self,
        credentials: Iterable[str],
        validity: int = 86400,
        cache: bool = True,  # does nothing
        urls: Optional[URLs] = None,
    ) -> None:
        """Initializes the AccessToken, but does not request a token.

        Parameters
        ----------
        - `credentials`: *(consumer_key, consumer_secret)*
            Authentication credentials in the form of a pair of key and secret.
        - `cache`: *bool, default: False*
            Set to false to always request new tokens, even if the current one has not expired.
        - `validity`: *int*
            Validity period for tokens, in seconds
        - `urls`: *URLs, optional*
            URLs instance to be used, will be initialized to the default if not provided.
        """
        self.credentials = Credentials(*credentials)
        self.validity_period = validity
        self.urls = urls or URLs()
        self.cache = cache

    def __str__(self) -> str:
        """Return the current token in str form."""
        return self.access_token

    @property
    def expiration(self) -> datetime:
        """Expiration of the current token string"""
        # Generate a token only when uninitialized
        if self._expiration == 0:
            self._update_token_data()
        return datetime.fromtimestamp(self._expiration)

    @property
    def access_token(self) -> str:
        """Token string"""
        expires_in = self._expiration - time.time()
        if expires_in > 0:
            logger.debug(f"Current token {self._access_token} expires in {expires_in} seconds.")
        else:
            # If we don't have a token, just get one
            logger.debug(f"Requesting new token")
            self._update_token_data()
            expires_in = self._expiration - time.time()

        # Renew token when if  there's less than request_margin time to expire
        if expires_in < self.request_margin:
            tries = 0
            previous_token = self._access_token
            logger.debug(
                f"Token expires in {expires_in:.2f}, starting renewal of {self._access_token}."
            )
            # Loop until we are sure we got the new token
            while (
                tries < 20
                and self._access_token == previous_token
                or expires_in < self.request_margin
            ):
                tries += 1
                time.sleep(0.5)
                logger.debug(f"Requesting new token...")
                self._update_token_data()
                logger.debug(f"Received/previous {self._access_token}/{previous_token})")
                expires_in = self._expiration - time.time()
            if tries >= 100:
                raise EumdacError(
                    f"Could not get fresh token from server, got {self._access_token}, which expires {datetime.fromtimestamp(self._expiration)}"
                )

        return self._access_token

    @property
    def auth(self) -> AuthBase:
        """Authentication object using the current token"""
        return HTTPBearerAuth(self.access_token)

    def _update_token_data(self) -> None:
        """Request a new token and renew the expiration time"""
        auth = HTTPBasicAuth(*self.credentials)
        now = time.time()
        response = post(
            self.urls.get("token", "token"),
            auth=auth,
            data={"grant_type": "client_credentials", "validity_period": self.validity_period},
            headers=eumdac.common.headers,
        )
        response.raise_for_status()
        token_data = response.json()
        self._expiration = now + token_data["expires_in"]
        self._access_token = token_data["access_token"]

    def _revoke(self) -> None:
        """Revoke the current token"""
        auth = HTTPBasicAuth(*self.credentials)
        response = post(
            self.urls.get("token", "revoke"),
            auth=auth,
            data={"grant_type": "client_credentials", "token": self._access_token},
            headers=eumdac.common.headers,
        )
        response.raise_for_status()
        self._expiration = 0
        self._access_token = ""


class AnonymousAccessToken(BaseToken):
    """Token class for anonymous access, provides no authentication parameters."""

    def __init__(self, urls: Optional[URLs] = None):
        """Init the token."""
        self.urls = urls or URLs()

    @property
    def auth(self) -> Optional[AuthBase]:
        """Return None"""
        return None



"""Module containing the Data Tailor class and related errors"""

from __future__ import annotations

import json
from typing import TYPE_CHECKING

from eumdac.customisation import Customisation
from eumdac.errors import EumdacError, eumdac_raise_for_status
from eumdac.tailor_models import Chain, DataTailorCRUD, Filter, Quicklook, RegionOfInterest
from eumdac.request import get, post
import eumdac.common
from eumdac.token import AccessToken, AnonymousAccessToken

if TYPE_CHECKING:  # pragma: no cover
    import sys
    from typing import Any, Optional

    from eumdac.product import Product
    from eumdac.token import BaseToken, URLs

    if sys.version_info < (3, 9):
        from typing import Iterable, Mapping, Sequence
    else:
        from collections.abc import Iterable, Mapping, Sequence


class DataTailor:
    """Interface with the EUMETSAT Data Tailor Webservice

    Instance it by providing a token and access the Data Tailor functions like
    posting new customisation jobs, listing the current jobs, cancelling running ones,
    download job outputs, and delete finished jobs.
    """

    token: BaseToken
    urls: URLs
    chains: DataTailorCRUD
    filters: DataTailorCRUD
    rois: DataTailorCRUD
    quicklooks: DataTailorCRUD
    _info: Optional[Mapping[str, Any]] = None
    _user_info: Optional[Mapping[str, Any]] = None

    def __init__(self, token: BaseToken) -> None:
        self.token = token
        self.urls = token.urls
        self.chains = DataTailorCRUD(self, Chain)
        self.filters = DataTailorCRUD(self, Filter)
        self.rois = DataTailorCRUD(self, RegionOfInterest)
        self.quicklooks = DataTailorCRUD(self, Quicklook)

    @property
    def customisations(self) -> Sequence[Customisation]:
        """Return the list of customisations"""
        url = self.urls.get("tailor", "customisations")
        response = get(
            url,
            auth=self.token.auth,
            headers=eumdac.common.headers,
        )
        eumdac_raise_for_status("Could not get customisations", response, DataTailorError)
        customisations = response.json()["data"]
        return [Customisation.from_properties(properties, self) for properties in customisations]

    @property
    def info(self) -> Mapping[str, Any]:
        """Return information about Data Tailor Webservice in a Dict-like format."""
        if self._info is None:
            url = self.urls.get("tailor", "info")
            auth = self.token.auth
            response = get(
                url,
                auth=auth,
                headers=eumdac.common.headers,
            )
            eumdac_raise_for_status("Could not get info", response, DataTailorError)
            self._info = response.json()
        return self._info

    @property
    def user_info(self) -> Mapping[str, Any]:
        """Return information about the current Data Tailor Webservice user in a Dict-like format."""
        if self._user_info is None:
            url = self.urls.get("tailor", "user info")
            auth = self.token.auth
            response = get(
                url,
                auth=auth,
                headers=eumdac.common.headers,
            )
            eumdac_raise_for_status("Could not get user_info", response, DataTailorError)
            self._user_info = response.json()
        return self._user_info

    @property
    def quota(self) -> Mapping[str, Any]:
        """Return information about the user workspace quota on the Data Tailor Webservice in a Dict-like format."""
        url = self.urls.get("tailor", "report quota")
        auth = self.token.auth
        response = get(
            url,
            auth=auth,
            headers=eumdac.common.headers,
        )
        eumdac_raise_for_status("Could not get quota", response, DataTailorError)
        return response.json()

    @property
    def is_local(self) -> bool:
        """Return if the configured Data Tailor is the Data Tailor Webservice or a local instance."""
        # when no token for datatailor exists we assume this is a local tailor instance
        return isinstance(self.token, AnonymousAccessToken)

    def get_customisation(self, cutomisation_id: str) -> Customisation:
        """Return a customisation job given its id"""
        return Customisation(cutomisation_id, self)

    def new_customisation(self, product: Product, chain: Chain) -> Customisation:
        """Start a new customisation job for the given product.

        Started customisations will run asynchronously and need to be monitored.
        Once finished, their outputs can be downloaded, and then they need to be deleted.

        Arguments
        ---------
        - `product` : *Product*
            Data Store product to customise
        - `chain` : *Chain*
            Chain configuration to use for the customisation
        """
        (customisation,) = self.new_customisations([product], chain)
        return customisation

    def new_customisations(
        self, products: Iterable[Product], chain: Chain
    ) -> Sequence[Customisation]:
        """Starts multiple customisation jobs for the given products

        Started customisations will run asynchronously and need to be monitored.
        Once finished, their outputs can be downloaded, and then they need to be deleted.

        Arguments
        ---------
        - `products` : *Iterable[Product]*
            Data Store products to customise
        - `chain` : *Chain*
            Chain configuration to use for the customisation
        """
        product_paths = "|||".join(
            self.urls.get(
                "datastore",
                "download product",
                vars={
                    "product_id": product._id,
                    "collection_id": product.collection._id,
                },
            )
            for product in products
        )

        data = {"product_paths": product_paths}
        params = {}

        # instead of guessing the correct token, datatailor should use the token attached to each product
        if isinstance(self.token, AccessToken):
            # provide own token to the endpoint since we assume it is valid for datastore
            params["access_token"] = str(self.token)
        elif self.is_local and any(products):
            # for local tailor instances we use the token attached to the first product
            params["access_token"] = str(next(iter(products)).datastore.token)
        if isinstance(chain, str):
            data["chain_name"] = chain
        else:
            data["chain_config"] = json.dumps(chain.asdict())

        response = post(
            self.urls.get("tailor", "customisations"),
            auth=self.token.auth,
            params=params,
            files=data,
            headers=eumdac.common.headers,
        )

        eumdac_raise_for_status("Could not add customizations", response, DataTailorError)

        customisation_ids = response.json()["data"]
        return [self.get_customisation(customisation_id) for customisation_id in customisation_ids]


class DataTailorError(EumdacError):
    """Errors related to DataTailor operations"""



"""
This module defines errors and error handling functions for eumdac.
"""

import json
from typing import *
from urllib.parse import urlparse

import requests


def eumdac_raise_for_status(
    msg: str, response: requests.Response, exc_cls: Type[Exception]
) -> None:
    """Raises an EumdacError with the given message wrapping an HTTPError, if one occurred.

    Raises
    ------
    - `EumdacError`
        If the provided response raises an HTTPError
    """

    try:
        response.raise_for_status()
    except requests.HTTPError as exc:
        url = urlparse(response.url)
        response_text = response.text
        if not response_text and response.raw:
            response_text = response.raw.data
        try:
            extra_info = json.loads(response_text)
        except json.decoder.JSONDecodeError:
            extra_info = {"text": response_text}
        extra_info.update({"url": url, "status": response.status_code})
        if response.status_code == 401:
            msg += " - Authentication error (401)"
        elif response.status_code == 403:
            msg += " - Unauthorised (403)"
        elif response.status_code == 404:
            msg += " - Not found (404)"
        if response.status_code > 500:
            msg += f" (due to a server-side error ({response.status_code})"
        exception = exc_cls(msg, extra_info)
        raise exception from exc


class EumdacError(Exception):
    """Common base class for eumdac errors

    Attributes
    ----------
    - `msg` : *str*
        exception text
    - `extra_info` : *Optional[Dict[str, Any]]*
        Dictionary containing additional information
        The title and description entries are embedded into the msg attribute, if present
    """

    def __init__(self, msg: str, extra_info: Optional[Dict[str, Any]] = None):
        """Init the error, putting common extra_info members into the message."""
        self.extra_info = extra_info
        if extra_info:
            if "title" in extra_info:
                msg = f"{msg} - {extra_info['title']}"
            if "description" in extra_info:
                msg = f"{msg}. {extra_info['description']}"
        super().__init__(msg)



import fnmatch
import io
import time
import shutil
import unittest
from datetime import datetime

from pytest import skip

from eumdac.token import AccessToken
from eumdac.datastore import DataStore
from eumdac.tailor_models import Chain
from eumdac.datatailor import DataTailor
from eumdac.customisation import Customisation, AlreadyDeletedCustomisationError

from .base import DataServiceTestCase, INTEGRATION_TESTING


class TestCustomisation(DataServiceTestCase):
    def setUp(self):
        super().setUp()
        self.token = AccessToken(self.credentials)
        self.datatailor = DataTailor(self.token)

    @unittest.skip("Temporarily skipping")
    def test_full_customisation_process(self):
        # format conversion
        chain_config = Chain(product="HRSEVIRI", format="netcdf4", quicklook="hrseviri_png")
        # search for the last product in June 2020
        datastore = DataStore(self.token)
        hrseviri = datastore.get_collection("EO:EUM:DAT:MSG:HRSEVIRI")
        product = hrseviri.search(dtend=datetime(2020, 7, 1)).first()
        # use context manager, to automatically delete customisations at the end even
        # if an error occurs
        with self.datatailor.new_customisation(product, chain_config) as customisation:
            if not INTEGRATION_TESTING:
                customisation.update_margin = 0
            timeout = 900  # needed to increase to 12 minutes, overstrained DT...
            tic = time.time()
            while time.time() - tic < timeout:
                if INTEGRATION_TESTING:
                    time.sleep(5)
                if customisation.status == "DONE":
                    break
            else:
                raise TimeoutError(f"Cutomisation took longer than {timeout}s")
            # test streaming
            (png_aux_xml,) = fnmatch.filter(customisation.outputs, "*.png.aux.xml")
            with customisation.stream_output(png_aux_xml) as stream:
                file = io.BytesIO()
                shutil.copyfileobj(stream, file)
            # check properties
            self.assertIsInstance(customisation.creation_time, datetime)
            self.assertIsInstance(customisation.backend, str)
            self.assertIsInstance(customisation.product_type, str)
            self.assertIsInstance(customisation.processing_steps, list)
            self.assertIsInstance(customisation.status, str)
            self.assertIsInstance(customisation.progress, int)
            self.assertIsInstance(customisation.duration, int)
            self.assertIsInstance(customisation.outputs, list)
            self.assertIsInstance(customisation.logfile, str)

        expected_xml_length = 870
        self.assertEqual(file.tell(), expected_xml_length)

        # check that customisation has been removed
        delete_url = self.datatailor.urls.get("tailor", "delete")
        self.requests_mock.assert_call_count(delete_url, 1)

    @unittest.skipIf(INTEGRATION_TESTING, "Integration already covered!")
    def test_properties_cache(self):
        customisation_id = "abcdef123"
        url = self.datatailor.urls.get(
            "tailor", "customisation", vars={"customisation_id": customisation_id}
        )
        self.requests_mock.add(
            "GET",
            url,
            json={
                customisation_id: {
                    "creation_time": datetime.now().strftime(Customisation._creation_time_format),
                    "backend_id": "backend-id",
                    "product_id": "product-id",
                    "required_processing_steps": ["a", "b", "c"],
                    "status": "DONE",
                    "progress": 100,
                    "processing_duration": 123,
                    "output_products": ["file.nc", "file.xml", "file.png"],
                }
            },
        )
        self.token._expiration = time.time() + 1000
        self.token._access_token = "token"
        customisation = Customisation(customisation_id, self.datatailor)

        creation_time = customisation.creation_time
        backend = customisation.backend
        self.assertIsInstance(creation_time, datetime)
        self.assertIsInstance(backend, str)
        self.requests_mock.assert_call_count(url, 1)

        # force reload
        customisation.update_margin = 0
        status = customisation.status
        self.assertIsInstance(status, str)
        self.requests_mock.assert_call_count(url, 2)

    @unittest.skipIf(INTEGRATION_TESTING, "Integration already covered!")
    def test_string_representation(self):
        customisation_id = "abcdef123"
        customisation = Customisation(customisation_id, self.datatailor)
        self.assertEqual(customisation_id, str(customisation))
        self.assertIn(customisation_id, repr(customisation))

    @unittest.skipIf(INTEGRATION_TESTING, "Test Failure")
    def test_cannot_fails_when_deleted(self):
        self.requests_mock.add("PATCH", self.datatailor.urls.get("tailor", "delete"))
        self.token._expiration = time.time() + 1000
        self.token._access_token = "token"
        customisation_id = "abcdef123"
        customisation = Customisation(customisation_id, self.datatailor)
        customisation.delete()

        message = "Customisation already deleted."
        with self.assertRaisesRegex(AlreadyDeletedCustomisationError, message):
            customisation.status

        with self.assertRaisesRegex(AlreadyDeletedCustomisationError, message):
            customisation.logfile

        with self.assertRaisesRegex(AlreadyDeletedCustomisationError, message):
            with customisation.stream_output("file.dat") as stream:
                data = stream.read()
                del data

    @unittest.skipIf(INTEGRATION_TESTING, "Test Failure")
    def test_fail_on_unknown_stream_output(self):
        customisation_id = "abcdef123"
        url = self.datatailor.urls.get(
            "tailor", "customisation", vars={"customisation_id": customisation_id}
        )
        self.requests_mock.add(
            "GET",
            url,
            json={customisation_id: {"output_products": ["file.nc", "file.xml", "file.png"]}},
        )
        self.token._expiration = time.time() + 1000
        self.token._access_token = "token"
        customisation_id = "abcdef123"
        customisation = Customisation(customisation_id, self.datatailor)
        message = r"file\.dat not in "
        with self.assertRaisesRegex(ValueError, message):
            with customisation.stream_output("file.dat") as stream:
                data = stream.read()
                del data

    @classmethod
    def prepare_integration_test(cls):
        prepare_test_customisation(cls.credentials)


def prepare_test_customisation(credentials):
    """delete all existing customisations"""
    token = AccessToken(credentials)
    datatailor = DataTailor(token)
    for customisation in datatailor.customisations:
        try:
            customisation.delete()
        except:
            pass



import unittest
import time
from unittest import mock

from requests.exceptions import HTTPError

from eumdac.errors import EumdacError
from eumdac.token import AccessToken
from eumdac.datatailor import DataTailor
from eumdac.customisation import Customisation
from eumdac.tailor_models import Chain, Filter, RegionOfInterest, Quicklook

from .base import DataServiceTestCase, INTEGRATION_TESTING


# @pytest.mark.skip("Temporarily skipping")
class TestDataTailor(DataServiceTestCase):
    def setUp(self):
        super().setUp()
        self.token = AccessToken(self.credentials)
        self.datatailor = DataTailor(self.token)

    @unittest.skipIf(INTEGRATION_TESTING, "Tested in unit testing phase")
    def test_property_customisations(self):
        token_url = self.token.urls.get("token", "token")
        self.requests_mock.add(
            "POST",
            token_url,
            json={
                "access_token": "b070adda-00c8-36cd-bb8f-4f6a2d6661ce",
                "scope": "am_application_scope default",
                "token_type": "Bearer",
                "expires_in": 970,
            },
        )
        customisations_url = self.datatailor.urls.get("tailor", "customisations")
        self.requests_mock.add(
            "GET",
            customisations_url,
            json={
                "data": [
                    {
                        "user_id": "ABC",
                        "creation_time": "20210810T142328Z",
                        "backend_id": "epct_gis_msg",
                        "product_id": "HRSEVIRI",
                        "required_processing_steps": ["format"],
                        "status": "DONE",
                        "progress": 100,
                        "processing_duration": 47,
                        "output_products": ["output.file"],
                        "id": "abcdef",
                    }
                ]
            },
        )
        customisations = self.datatailor.customisations
        self.assertIsInstance(customisations, list)
        self.assertIsInstance(customisations[0], Customisation)

    def test_properties(self):
        info = self.datatailor.info
        self.assertIn("epct_version", info)

        user_info = self.datatailor.user_info
        self.assertIsInstance(user_info["username"], str)

        quota = self.datatailor.quota
        self.assertIsInstance(quota, dict)

    @unittest.skipIf(INTEGRATION_TESTING, "Only checking the interface.")
    def test_create_customisation(self):
        self.token._access_token = "b070adda-00c8-36cd-bb8f-4f6a2d6661ce"
        self.token._expiration = time.time() + 1000
        self.requests_mock.add(
            "POST",
            self.datatailor.urls.get("tailor", "customisations"),
            json={"total": 1, "data": ["63f53c99"]},
        )
        chain_name = "MyChain"
        mock_product = mock.MagicMock(name="product")
        mock_product._id = "MY-PRODUCT-ID"
        mock_product.collection._id = "MY-COLLECTION-ID"
        customisation = self.datatailor.new_customisation(mock_product, chain_name)
        self.assertIsInstance(customisation, Customisation)

    def test_chain_crud(self):
        url = self.datatailor.chains.url
        chain = Chain(
            id="hrseviri_native_to_netcdf4",
            name="Native to netCDF4",
            description="Convert a SEVIRI Native product to netCDF4",
            product="HRSEVIRI",
            format="netcdf4",
        )
        # create
        self.datatailor.chains.create(chain)
        self.requests_mock.assert_call_count(url, 1)

        # read
        chain_ = self.datatailor.chains.read(chain.id)
        self.assertEqual(chain_, chain)
        chains = self.datatailor.chains.search(product="HRSEVIRI")
        self.assertIn(chain, chains)
        self.assertIsInstance(chains[0], Chain)

        # update
        chain.description = "convert a SEVIRI native product to netCDF4"
        self.datatailor.chains.update(chain)
        chain_ = self.datatailor.chains.read(chain.id)
        self.assertEqual(chain_, chain)

        # delete
        self.datatailor.chains.delete(chain)
        self.requests_mock.assert_call_count(url + f"/{chain.id}", 4)

    def test_filter_crud(self):
        url = self.datatailor.filters.url
        fltr = Filter(
            id="hrseviri_water_vapour",
            product="HRSEVIRI",
            bands=["channel_9", "channel_10"],
            name="Water Vapour",
        )
        # create
        self.datatailor.filters.create(fltr)
        self.requests_mock.assert_call_count(url, 1)

        # read
        fltr_ = self.datatailor.filters.read(fltr.id)
        self.assertEqual(fltr_, fltr)
        fltrs = self.datatailor.filters.search(product="HRSEVIRI")
        self.assertIn(fltr, fltrs)
        self.assertIsInstance(fltrs[0], Filter)

        # update
        fltr.bands = ["channel_9"]
        self.datatailor.filters.update(fltr)
        fltr_ = self.datatailor.filters.read(fltr.id)
        self.assertEqual(fltr_, fltr)

        # delete
        self.datatailor.filters.delete(fltr)
        self.requests_mock.assert_call_count(url + f"/{fltr.id}", 4)

    def test_roi_crud(self):
        url = self.datatailor.rois.url
        roi = RegionOfInterest(
            id="deutschland", name="Deutschland", NSWE=[54.983104, 47.302488, 5.988658, 15.016996]
        )
        # create
        self.datatailor.rois.create(roi)
        self.requests_mock.assert_call_count(url, 1)

        # read
        roi_ = self.datatailor.rois.read(roi.id)
        self.assertEqual(roi_, roi)

        # delete
        self.datatailor.rois.delete(roi)
        self.requests_mock.assert_call_count(url + f"/{roi.id}", 2)

    def test_quicklook_crud(self):
        url = self.datatailor.quicklooks.url
        quicklook = Quicklook(
            id="hrseviri_wv_jpeg",
            name="HRSEVIRI Water Vapour",
            resample_method="cubic",
            stretch_method="min_max",
            product="HRSEVIRI",
            format="jpeg_rgb",
            filter=Filter(bands=["channel_11", "channel_10", "channel_9"]),
        )
        # create
        self.datatailor.quicklooks.create(quicklook)
        self.requests_mock.assert_call_count(url, 1)

        # read
        quicklook_ = self.datatailor.quicklooks.read(quicklook.id)
        self.assertEqual(quicklook_, quicklook)
        quicklooks = self.datatailor.quicklooks.search(product="HRSEVIRI")
        self.assertIn(quicklook, quicklooks)
        self.assertIsInstance(quicklooks[0], Quicklook)

        # update
        quicklook.resample_method = "bilinear"
        self.datatailor.quicklooks.update(quicklook)
        quicklook_ = self.datatailor.quicklooks.read(quicklook.id)
        self.assertEqual(quicklook_, quicklook)

        # delete (using string)
        self.datatailor.quicklooks.delete(quicklook.id)
        self.requests_mock.assert_call_count(url + f"/{quicklook.id}", 4)

    @classmethod
    def prepare_integration_test(cls):
        prepare_test_datatailor(cls.credentials)


def prepare_test_datatailor(credentials):
    """delete any resource that will be created during the test"""
    token = AccessToken(credentials)
    datatailor = DataTailor(token)
    crud_resources = [
        (datatailor.chains, "hrseviri_native_to_netcdf4"),
        (datatailor.filters, "hrseviri_water_vapour"),
        (datatailor.rois, "deutschland"),
        (datatailor.quicklooks, "hrseviri_wv_jpeg"),
    ]
    for api, resource_id in crud_resources:
        try:
            api.delete(resource_id)
        except:
            # Resources may or may not be there and DTWS won't send 40X error codes
            # when deleting not-found resources, so the fastest way is to try
            # deleting and ignore errors
            pass



from eumdac import __version__ as v

dotversion = v.split("-")[0]
dotversion = dotversion.split("rc")[0]
commaversion = dotversion.replace(".", ", ") + ", 0"

winVersionFile = open("eumdac-win-version.py", "r")
data = winVersionFile.read()
data = data.replace("@@VERSION@@", dotversion)
data = data.replace("@@COMMAVERSION@@", commaversion)
winVersionFile.close()

output = open("eumdac-win-version.py", "w")
output.write(data)
output.close()



import re
from eumdac.logging import logger
from typing import Tuple, List


def is_collection_valid_for_coverage(collection: str) -> bool:
    cs = [
        "0662",
        "0665",
        "0672",
    ]
    for c in cs:
        if re.match(rf"EO\:EUM(IVV|VAL)?\:DAT\:{c}(:COM)?", collection):
            return True
    return False


def build_entries_from_coverage(coverage: str) -> Tuple[List[str], int]:
    entries = []
    expected = -1

    # Using if-elif to be python < 3.10 compliant
    if coverage == "FD":
        logger.info("Downloading all chunks inside the full disk: 01-40")
        entries.extend(["*_????_00[0-3][0-9].nc", "*_????_0040.nc"])
        expected = 40
    elif coverage == "H1":
        logger.info("Downloading chunks inside H1: 01-21")
        entries.extend(["*_????_000[1-9].nc", "*_????_001[0-9].nc", "*_????_002[0-1].nc"])
        expected = 21
    elif coverage == "H2":
        logger.info("Downloading chunks inside H2: 20-40")
        entries.extend(["*_????_002[0-9].nc", "*_????_003[0-9].nc", "*_????_0040.nc"])
        expected = 21
    elif coverage == "T1":
        logger.info("Downloading chunks inside T1: 01-16")
        entries.extend(["*_????_000[1-9].nc", "*_????_001[0-6].nc"])
        expected = 16
    elif coverage == "T2":
        logger.info("Downloading chunks inside T2: 13-27")
        entries.extend(["*_????_001[3-9].nc", "*_????_002[0-7].nc"])
        expected = 15
    elif coverage == "T3":
        logger.info("Downloading chunks inside T3: 26-40")
        entries.extend(["*_????_002[6-9].nc", "*_????_003[0-9].nc", "*_????_0040.nc"])
        expected = 15
    elif coverage == "Q1":
        logger.info("Downloading chunks inside Q1: 01-13")
        entries.extend(["*_????_000[0-9].nc", "*_????_001[0-3].nc"])
        expected = 13
    elif coverage == "Q2":
        logger.info("Downloading chunks inside Q2: 10-21")
        entries.extend(["*_????_001[0-9].nc", "*_????_002[0-1].nc"])
        expected = 12
    elif coverage == "Q3":
        logger.info("Downloading chunks inside Q3: 20-30")
        entries.extend(["*_????_002[0-9].nc", "*_????_0030.nc"])
        expected = 11
    elif coverage == "Q4":
        logger.info("Downloading chunks inside Q4: 29-40")
        entries.extend(["*_????_0029.nc", "*_????_003[0-9].nc", "*_????_0040.nc"])
        expected = 12

    # Include TRAIL file (chunk 41) in all areas
    entries.append("*_????_0041.nc")
    expected += 1

    return (entries, expected)


# Removes subdirectories for printing entries
def pretty_print_entry(entry: str) -> str:
    if entry.find("/") > -1:
        return entry.split("/")[1]
    else:
        return entry



import pytest
import tempfile
import os

from datetime import timedelta
from eumdac.lockfile import open_locked


@pytest.fixture(scope="function")
def lockfile_path(tmp_path):
    lockfile = tmp_path / "lockfile"
    yield lockfile
    if os.path.exists(lockfile):
        os.remove(lockfile)


def test_open_lock(lockfile_path):
    with open_locked(lockfile_path) as lf:
        assert os.path.exists(lockfile_path)


def test_multiple_lock(lockfile_path):
    with open_locked(lockfile_path) as lf:
        with open_locked(lockfile_path, timeout=timedelta(seconds=1)) as lf2:
            assert lf2 is None



import os
import unittest
from io import StringIO
from itertools import islice
from pathlib import Path
from shutil import rmtree
from unittest.mock import mock_open, patch

import pytest

from eumdac.cli import cli

from .base import INTEGRATION_TESTING


@pytest.fixture
@unittest.skipIf(INTEGRATION_TESTING, "Covered by integration testing")
def temp_output_dir(request):
    output_dir_name = f"output_dir_{request.node.name}"
    yield Path(output_dir_name)
    rmtree(output_dir_name)


@pytest.fixture
@unittest.skipIf(INTEGRATION_TESTING, "Covered by integration testing")
def temp_config_dir(request):
    config_dir_name = Path(f"config_dir_{request.node.name}")
    os.environ["EUMDAC_CONFIG_DIR"] = str(config_dir_name)
    config_dir_name.mkdir(exist_ok=True)
    yield config_dir_name
    rmtree(config_dir_name)
    del os.environ["EUMDAC_CONFIG_DIR"]


@pytest.fixture
@unittest.skipIf(INTEGRATION_TESTING, "Covered by integration testing")
def temp_credentials(temp_config_dir):
    credentials_path = temp_config_dir / "credentials"
    with credentials_path.open("w") as fobj:
        fobj.write("user,password")
    yield credentials_path


def eumdac(args):
    test_out = StringIO()
    test_err = StringIO()

    with patch("sys.stdout", test_out), patch("sys.stderr", test_err), patch(
        "eumdac.cli.sys.stdin", new_callable=mock_open()
    ) as mock_stdin:
        mock_stdin.isatty.return_value = False
        try:
            cli(args)
        except SystemExit as exc:
            if exc.code != 0:
                raise exc

    return test_out.getvalue(), test_err.getvalue()


def assert_eumdac_output(args, expected_lines):
    out, err = eumdac(args)
    print(out)
    expected_block = "\n".join(str(x) for x in expected_lines)
    error_msg = "\n{}\nvs\n{}".format(expected_block, out)
    assert_list_contains(out.splitlines(), expected_lines, error_msg)
    assert not err, "stderr not empty!\n" "Contents:\n" f"{err}"


def assert_list_contains(outlist, expectedlist, msg=None):
    iout = iter(outlist)
    iexpected = iter(expectedlist)
    for exp in iexpected:
        print(exp)
        if isinstance(exp, tuple) and exp[0] == "order_does_not_matter":
            inner_expectation = sorted(exp[1])
            out_slice = sorted(islice(iout, len(exp[1])))
            print(inner_expectation)
            print(out_slice)
            assert_list_contains(out_slice, inner_expectation)
        else:
            out = next(iout)
            if msg:
                assert exp in out, msg
            else:
                assert exp in out


def test_assert_list_contains():
    assert_list_contains(["A1", "A2", "A3", "B1"], ["A", "A", "A", "B"])


def test_assert_list_contains_simple_assertion_error():
    with pytest.raises(AssertionError):
        assert_list_contains(["A1", "A2", "A3", "B1"], ["A", "B", "A", "B"])


def test_assert_list_contains_assertion_error_unequal_len():
    with pytest.raises(AssertionError):
        assert_list_contains(["A1", "A2", "A3"], ["A", "B", "A", "B"])


def test_assert_list_contains_with_indifferent_order():
    assert_list_contains(
        ["A1", "A2", "A3", "B1"], ["A", ("order_does_not_matter", ["B", "A", "A"])]
    )


@unittest.skipIf(INTEGRATION_TESTING, "Covered by unit testing")
def test_set_credentials(temp_config_dir):
    args = "--set-credentials user password".split()
    expected = (
        "Credentials are correct. Token was generated:",
        "Credentials are written to file",
    )
    with patch("eumdac.token.AccessToken.access_token", "TOKEN"):
        assert_eumdac_output(args, expected)


@unittest.skipIf(INTEGRATION_TESTING, "Covered by unit testing")
def test_download_product(temp_credentials, temp_output_dir):
    args = f"download -c MockCollection -p MockProduct -o {temp_output_dir} --test".split()
    expected = (
        "Processing",
        "Using order",
        "Output directory:",
        "Preparing download",
        "Download complete: MockProduct",
        "Removing successfully finished order",
    )
    assert_eumdac_output(args, expected)


@unittest.skipIf(INTEGRATION_TESTING, "Covered by unit testing")
def test_skip_download_product(temp_credentials, temp_output_dir):
    args = f"download -c MockCollection -p MockProduct -o {temp_output_dir} --test".split()
    expected = ("Processing", "Using order", "Output directory:", "Skip")

    # First run to ensure the file exists already
    eumdac(args)
    assert_eumdac_output(args, expected)


@unittest.skipIf(INTEGRATION_TESTING, "Covered by unit testing")
def test_download_product_entry(temp_credentials, temp_output_dir):
    args = f"download -c MockCollection -p MockProduct --entry *.nc -o {temp_output_dir} --test".split()
    # we expect a folder with the product name to be created when --entry is given
    expected = (
        "Processing",
        "Using order",
        "Output directory:",
        (
            "order_does_not_matter",
            [
                "Job 1: Preparing download",
                "Job 1: Preparing download",
                "Job 1: Download complete",
                "Job 1: Download complete",
            ],
        ),
        "Removing successfully finished order",
    )
    assert_eumdac_output(args, expected)


@unittest.skipIf(INTEGRATION_TESTING, "Covered by unit testing")
def test_download_output_dir(temp_credentials, temp_output_dir):
    args = f"download -c MockCollection --time-range 2020-03-01 2020-03-01T12:15 -o {temp_output_dir} --test".split()
    expected = (
        "Processing",
        "Using order",
        "Output directory:",
        (
            "order_does_not_matter",
            [
                "Job 1: Preparing download",
                "Job 2: Preparing download",
                "Job 1: Download complete",
                "Job 2: Download complete",
            ],
        ),
        "Removing successfully finished order",
    )
    assert_eumdac_output(args, expected)



from pathlib import Path
from eumdac.order import (
    Order,
)
import concurrent.futures

from tests.base import FakeProduct, FakeCollection


def update_order(order, cid, pid, status, download_state):
    if download_state:
        download_states = {pid: download_state}
    else:
        download_states = None
    order.update(
        customisation_id=cid, product_id=pid, status=status, download_states=download_states
    )


def test_order_concurrency(tmp_path):
    products = []
    for i in range(0, 2):
        products.append(FakeProduct(id=str(i), collection=FakeCollection(str(i))))

    for repetitions in range(0, 100):
        o = Order(tmp_path / str(repetitions))
        o.initialize(
            chain=None, products=products, output_dir=Path("."), file_pattern=None, query=None
        )

        for p in products:
            o.update(
                customisation_id=p.collection._id,
                product_id=p._id,
                status="INIT",
                download_states={p._id: "INIT"},
            )

        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
            f_res1 = [
                executor.submit(update_order, o, p.collection._id, p._id, "MOD", None)
                for p in products
            ]
            concurrent.futures.wait(
                f_res1, timeout=None, return_when=concurrent.futures.ALL_COMPLETED
            )
            f_res2 = [
                executor.submit(update_order, o, p.collection._id, p._id, None, "MOD")
                for p in products
            ]
            concurrent.futures.wait(
                f_res2, timeout=None, return_when=concurrent.futures.ALL_COMPLETED
            )

        with o.dict_from_file() as order_d:
            for product_id in order_d["products_to_process"].keys():
                ss = order_d["products_to_process"][product_id]["server_state"]

                if "download_states" in order_d["products_to_process"][product_id]:
                    download_states = order_d["products_to_process"][product_id]["download_states"]
                    dl_complete = [
                        True if ds == "MOD" else False for ds in download_states.values()
                    ]
                    assert all(dl_complete)
                    assert ss == "MOD"



VSVersionInfo(
  ffi=FixedFileInfo(
    filevers=(@@COMMAVERSION@@),
    prodvers=(@@COMMAVERSION@@),
    mask=0x3f,
    flags=0x0,
    OS=0x40004,
    fileType=0x1,
    subtype=0x0,
    date=(0, 0)
    ),
  kids=[
    StringFileInfo(
      [
      StringTable(
        u'040904B0',
        [StringStruct(u'CompanyName', u'EUMETSAT'),
        StringStruct(u'FileDescription', u'EUMETSAT Data Access Client'),
        StringStruct(u'FileVersion', u'@@VERSION@@'),
        StringStruct(u'InternalName', u'eumdac'),
        StringStruct(u'LegalCopyright', u'\xa9 EUMETSAT. All rights reserved.'),
        StringStruct(u'OriginalFilename', u'eumdac.exe'),
        StringStruct(u'ProductName', u'EUMETSAT Data Access Client'),
        StringStruct(u'ProductVersion', u'@@VERSION@@')])
      ]), 
    VarFileInfo([VarStruct(u'Translation', [0x0809, 1252])])
  ]
)


"""Module providing the JobIdentifier, a helper class for consistent identification and logging of activities."""

import uuid
from threading import Lock
from typing import Any, Dict, Tuple

from eumdac.errors import EumdacError


class JobIdentifier:
    """Wraps an activity as an identified job."""

    def __init__(self, total_jobs: int):
        """Init considering the expected `total_jobs`."""
        self.current_count = 0
        self.total_jobs = total_jobs
        self._lock = Lock()
        self.registered_objects: Dict[Any, Tuple[int, str]] = {}

    def register(self, obj: Any) -> None:
        """Register a new job from `obj`."""
        if obj in self.registered_objects:
            raise JobIdError(f"Object '{obj}' already registered.")
        self.registered_objects[obj] = (self._make_new_job_id(), str(uuid.uuid4()))

    def job_id_tuple(self, obj: Any) -> Tuple[int, str]:
        """Return a tuple that identifies the job for `obj`, if any."""
        try:
            return self.registered_objects[obj]
        except KeyError:
            raise JobIdError(
                f"No Job ID for '{obj}'. Available ones: {list(self.registered_objects.keys())}"
            )

    def job_id_str(self, obj: Any) -> str:
        return f"Job {self.job_id_tuple(obj)[0]}"

    def _make_new_job_id(self) -> int:
        """Reserve a new job id, if the total has not been reached."""
        with self._lock:
            self.current_count += 1
            if self.current_count > self.total_jobs:
                raise JobIdError(
                    "Too many Job IDs requested. "
                    f"Expected a maximum of {self.total_jobs} Job ID requests"
                )
            return self.current_count


class JobIdError(EumdacError):
    """JobIdentifier related errors."""

    pass



"""This module contains the eumdac requests wrapper that includes automatic retries and management of throttling."""

import json
import random
import time
from datetime import datetime
from typing import Any, Dict

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from eumdac.errors import EumdacError
from eumdac.logging import logger


class RequestError(EumdacError):
    """Error related to requests."""

    pass


class RetryAndLog(Retry):
    """Retry configuration that will log retry attempts.

    Extends urllib3.util.retry.Retry, decorating the 'increment' method.
    """

    def increment(  # type: ignore
        self,
        method: Any = None,
        url: Any = None,
        response: Any = None,
        error: Any = None,
        _pool: Any = None,
        _stacktrace: Any = None,
    ) -> Retry:
        """Decorated urllib3.util.retry.Retry::increment to include logging."""
        target_uri = ""
        if _pool:
            target_uri = f"{method} {_pool.scheme}://{_pool.host}:{_pool.port}{url}"
        elif error:
            target_uri = f"{error.conn.host}{url}"

        cause = ""
        if response and response.data:
            cause = f'server response {response.status} - "{response.data}" '
        if error:
            cause = f'{cause}error: "{error}"'

        logger.info(f"Trying again for {target_uri} due to {cause}")
        return super().increment(method, url, response, error, _pool, _stacktrace)


def _get_adapter(max_retries: int, backoff_factor: float) -> HTTPAdapter:
    """Prepare an an HTTPAdapter that will retry failed requests up to 'max_retries' times.

    Only requests that return a 50X error code will be retried.

    Parameters
    ----------
    - `max_retries` : *int*
        Number of retries to perform.
    - `backoff_factor` : *float*
        Backoff factor to apply between attempts after the second try.

    Returns
    -------
    - `HTTPAdapter`
        Adapter prepared with the given 'max_retries' and 'backoff_factor'.
    """

    retry = RetryAndLog(
        total=max_retries,
        backoff_factor=backoff_factor,
        status_forcelist=[500, 502, 503, 504],
        allowed_methods=["HEAD", "GET", "OPTIONS", "POST", "PUT", "PATCH"],
        raise_on_status=False,
    )

    return HTTPAdapter(max_retries=retry)


def _should_retry(response: requests.Response, backoff: int = random.randint(1, 6) * 10) -> bool:
    """Decide whether the request should be retried considering the received response.

    Handles the throttling of requests done by the Data Store API.

    Parameters
    ---------
    - `response` : *requests.Response*
        Response received from the server.
    - `backoff` : *int, optional*
        Backoff, in seconds, to apply between attempts, defaults to a random value smaller than 1 minute.
    """
    if response.status_code == 429:
        rd = json.loads(response.text)
        # handle throttling
        message = rd["message"]["reason"]
        if "message" in rd and "retryAfter" in rd["message"]:
            # Traffic limits exceeded
            timestamp = int(rd["message"]["retryAfter"]) / 1000
            utc_endtime = datetime.utcfromtimestamp(timestamp)
            duration = utc_endtime - datetime.utcnow()
            if duration.total_seconds() > 0:
                logger.warning(f"{rd['message']}: operation will resume in {duration}")
                time.sleep(duration.total_seconds())
                return True
        elif "message" in rd and "reason" in rd["message"]:
            if rd["message"]["reason"] == "Maximum number of connections exceeded":
                # Maximum number of connections exceeded
                logger.warning(f"{message}: throttling for {backoff}s")
                time.sleep(backoff)
                return True
            elif rd["message"]["reason"] == "Maximum number of requests exceeded":
                # Maximum number of requests exceeded
                logger.warning(f"{message}: throttling for 1s")
                time.sleep(1)
                return True

    return False


def _request(
    method: str,
    url: str,
    max_retries: int = 3,
    backoff_factor: float = 0.3,
    **kwargs: Any,
) -> requests.Response:
    """Perform a request with the given `method`, `url` and parameters with automatic retries and throttling management.

    Parameters
    ----------
    - `method`: *{'get', 'post', 'patch', 'put', 'delete'}*
        HTTP request method to use in the request.
    - `url`: *str*
        URL to make the request to.
    - `max_retries`: *int, optional*
        Max number of retries to perform if the request fails, default: 3.
    - `backoff_factor`: *float, optional*
        Backoff factor to apply between attempts, default 0.3.
    - `**kwargs`: *dict, optional*
        Extra arguments to pass to the request, refer to the requests library documentation for a list of possible arguments.

    Returns
    -------
    - `requests.Response`:
        Response received from the server.
    """

    adapter = _get_adapter(max_retries, backoff_factor)
    session = requests.Session()

    session.mount("http://", adapter)
    session.mount("https://", adapter)
    response = requests.Response()
    try:
        while True:
            if hasattr(session, method):
                logger.debug(_pretty_print(method, url, kwargs))
                response = getattr(session, method.lower())(url, **kwargs)
                if _should_retry(response):
                    continue
            else:
                raise RequestError(f"Operation not supported: {method}")
            break
    except (ValueError, KeyError, TypeError) as e:
        logger.error(f"Received unexpected response: {e}")
    except requests.exceptions.RetryError as e:
        raise RequestError(
            f"Maximum retries ({max_retries}) reached for {method.capitalize()} {url}"
        )

    return response


def get(url: str, **kwargs: Any) -> requests.Response:
    """Perform a GET HTTP request to the given `url` with the given parameters.

    Retries and throttling will be managed in a transparent way when making the request.

    Arguments
    ---------
    - `url`: *str*
        URL to make the request to.
    - `**kwargs`: *dict, optional*
        Extra arguments to pass to the request, refer to the requests library documentation for a list of possible arguments.

    Returns
    -------
    - `request.Response`:
        Response received from the server.
    """

    return _request("get", url, **kwargs)


def post(url: str, **kwargs: Any) -> requests.Response:
    """Perform a POST HTTP request to the given `url` with the given parameters.

    Retries and throttling will be managed in a transparent way when making the request.

    Arguments
    ---------
    - `url`: *str*
        URL to make the request to.
    - `**kwargs`: *dict, optional*
        Extra arguments to pass to the request, refer to the requests library documentation for a list of possible arguments.

    Returns
    -------
    - `request.Response`:
        Response received from the server.
    """
    return _request("post", url, **kwargs)


def patch(url: str, **kwargs: Any) -> requests.Response:
    """Perform a PATCH HTTP request to the given `url` with the given parameters.

    Retries and throttling will be managed in a transparent way when making the request.

    Arguments
    ---------
    - `url`: *str*
        URL to make the request to.
    - `**kwargs`: *dict, optional*
        Extra arguments to pass to the request, refer to the requests library documentation for a list of possible arguments.

    Returns
    -------
    - `request.Response`:
        Response received from the server.
    """
    return _request("patch", url, **kwargs)


def put(url: str, **kwargs: Any) -> requests.Response:
    """Perform a PUT HTTP request to the given `url` with the given parameters.

    Retries and throttling will be managed in a transparent way when making the request.

    Arguments
    ---------
    - `url`: *str*
        URL to make the request to.
    - `**kwargs`: *dict, optional*
        Extra arguments to pass to the request, refer to the requests library documentation for a list of possible arguments.

    Returns
    -------
    - `request.Response`:
        Response received from the server.
    """
    return _request("put", url, **kwargs)


def delete(url: str, **kwargs: Any) -> requests.Response:
    """Perform a DELETE HTTP request to the given `url` with the given parameters.

    Retries and throttling will be managed in a transparent way when making the request.

    Arguments
    ---------
    - `url`: *str*
        URL to make the request to.
    - `**kwargs`: *dict, optional*
        Extra arguments to pass to the request, refer to the requests library documentation for a list of possible arguments.

    Returns
    -------
    - `request.Response`:
        Response received from the server.
    """

    return _request("delete", url, **kwargs)


def _pretty_print(method: str, url: str, kwargs: Dict[str, Any]) -> str:
    """Returns a readable str of the given request."""
    pargs = {}
    for key in kwargs.keys():
        if key == "headers":
            headers = {}
            for header in kwargs[key]:
                if header not in ["referer", "User-Agent"]:
                    headers[header] = kwargs[key][header]
            if len(headers) > 0:
                pargs[key] = headers
        elif key == "auth":
            if hasattr(kwargs[key], "token"):
                pargs[key] = f"Bearer {str(kwargs[key].token)}"  # type: ignore
            else:
                pargs[key] = f"{type(kwargs[key]).__name__}"  # type: ignore
        else:
            pargs[key] = kwargs[key]
    return f"Request: {method.upper()} {url}, payload: {pargs}"



import pytest
from eumdac.job_id import JobIdentifier, JobIdError


def test_register():
    uut = JobIdentifier(3)

    uut.register("a")
    assert 1 == len(uut.registered_objects)
    uut.register("b")
    assert 2 == len(uut.registered_objects)
    uut.register("c")
    assert 3 == len(uut.registered_objects)


def test_register_raise_register_same_object_twice():
    uut = JobIdentifier(3)
    uut.register("a")
    with pytest.raises(JobIdError) as exc_info:
        uut.register("a")
    assert str(exc_info.value) == "Object 'a' already registered."


def test_total_jobs_exceed():
    uut = JobIdentifier(1)
    uut.register("a")
    with pytest.raises(JobIdError) as exc_info:
        uut.register("b")
    assert (
        str(exc_info.value) == "Too many Job IDs requested. Expected a maximum of 1 Job ID requests"
    )


def test_job_id_tuple():
    uut = JobIdentifier(1)
    uut.register("a")
    out = uut.job_id_tuple("a")
    assert out[0] == 1


def test_request_non_existent_object():
    uut = JobIdentifier(1)
    uut.register("a")
    with pytest.raises(JobIdError) as exc_info:
        uut.job_id_tuple("b")
    assert str(exc_info.value) == "No Job ID for 'b'. Available ones: ['a']"



"""Module that enables order management in eumdac CLI."""

import os
import re
import threading
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import datetime, timedelta
from pathlib import Path
from typing import *

import yaml

from eumdac.config import get_config_dir
from eumdac.datatailor import DataTailor
from eumdac.errors import EumdacError
from eumdac.lockfile import open_locked
from eumdac.logging import gen_table_printer
from eumdac.product import Product
from eumdac.tailor_models import Chain


@dataclass
class ProductInfo:
    p_id: str
    p_dict: Dict[str, Any]


class Order:
    def __init__(self, order_file: Optional[Path] = None, order_dir: Optional[Path] = None):
        if order_dir is None:
            order_dir = get_default_order_dir()
        self._order_file = order_file or new_order_filename(order_dir)
        self._lock = threading.Lock()
        self._update_lock = threading.Lock()

    def initialize(
        self,
        chain: Optional[Chain],
        products: Iterable[Product],
        output_dir: Path,
        file_pattern: Optional[Iterable[str]],
        query: Optional[Dict[str, str]],
        dirs: bool = False,
        onedir: bool = False,
        no_warning_logs: bool = True,
    ) -> None:
        self._chain = chain
        self._output_dir = output_dir
        self._file_patterns = file_pattern
        self._no_warning_logs = no_warning_logs

        order_info: Dict[str, Any] = {
            "file_patterns": file_pattern,
            "output_dir": str(output_dir.resolve()),
            "query": query,
            "dirs": dirs,
            "onedir": onedir,
        }

        if chain:
            order_info["type"] = "tailor"
            order_info["chain"] = chain.asdict()
            order_info["products_to_process"] = {
                p._id: {
                    "col_id": p.collection._id,
                    "server_state": "UNSUBMITTED",
                    "customisation": None,
                }
                for p in products
            }
        else:
            order_info["type"] = "download"
            order_info["products_to_process"] = {
                p._id: {
                    "col_id": p.collection._id,
                    "server_state": "UNSUBMITTED",
                }
                for p in products
            }

        with self._lock:
            with self._order_file.open("w") as orf:
                yaml.dump(
                    order_info,
                    orf,
                )

    def __str__(self) -> str:
        return Path(self._order_file).stem

    def get_dict_entries(self, *args: str) -> Tuple[Optional[str], ...]:
        ret: List[Optional[str]] = []
        with self.dict_from_file() as order_d:
            for name in args:
                try:
                    ret.append(order_d[name])
                except KeyError:
                    ret.append(None)
        return tuple(ret)

    def status(self) -> str:
        if self.all_done():
            return "DONE"
        for p_info in self.iter_product_info():
            if p_info.p_dict["server_state"] in ("FAILED", "INACTIVE"):
                return "FAILED"
        return "NOT COMPLETED"

    def delete(self) -> None:
        os.remove(self._order_file)

    def collections(self) -> List[str]:
        ret = []
        for p_info in self.iter_product_info():
            ret.append(p_info.p_dict["col_id"])
        return list(set(ret))

    def pretty_string(self, print_products: bool = False) -> str:
        ret_lines: List[str] = []
        (typ, query, chain, output_dir) = self.get_dict_entries(
            "type", "query", "chain", "output_dir"
        )
        ret = [
            f"Order {str(self)}",
            f"Status: {self.status()}",
            f"Collection: {self.collections()}",
            "Query:",
        ]
        query_dump = yaml.dump(query).strip()
        for line in query_dump.split("\n"):
            ret.append(f"    {line.rstrip()}")
        if chain:
            ret.append("Chain:")
            chain_dump = yaml.dump(chain).strip()
            for line in chain_dump.split("\n"):
                ret.append(f"    {line.rstrip()}")
        ret.append(f"Output directory: {output_dir}")
        if print_products:
            print_func = ret.append
            if typ == "tailor":
                printer = gen_table_printer(
                    print_func, [("Product", 60), ("Job Id", 10), ("Status", 12)]
                )
                for p_info in self.iter_product_info():
                    state = _compute_state(p_info.p_dict)
                    printer([p_info.p_id, p_info.p_dict["customisation"], state])
            elif typ == "download":
                printer = gen_table_printer(print_func, [("Product", 60), ("Status", 12)])
                for p_info in self.iter_product_info():
                    printer([p_info.p_id, p_info.p_dict["server_state"]])
            else:
                raise NotImplementedError(typ)
        return "\n".join(ret)

    def _locked_serialize(self, order_dict: Dict[str, Any]) -> None:
        with self._lock:
            with self._order_file.open("w") as orf:
                yaml.dump(order_dict, orf)

    @contextmanager
    def dict_from_file(self) -> Generator[Dict[str, Any], None, None]:
        with self._lock:
            ret_dict = self._deserialize()
        yield ret_dict

    def _deserialize(self) -> Dict[str, Any]:
        with self._order_file.open("r") as orf:
            ret_val = yaml.safe_load(orf)
        if ret_val is None:
            raise EumdacError(f"{self._order_file.resolve()} is corrupted.")
        return ret_val

    def remote_delete_failed(self, datatailor: DataTailor) -> None:
        for p_info in self.iter_product_info():
            if p_info.p_dict["server_state"] == "FAILED":
                customisation_id = p_info.p_dict["customisation"]
                if customisation_id:
                    try:
                        customisation = datatailor.get_customisation(customisation_id)
                        customisation.delete()
                    except EumdacError:
                        continue

    def resolve_product_num(self, product_id: str) -> Tuple[int, int]:
        num_products = len(list(self.iter_product_info()))
        for num, p_info in enumerate(self.iter_product_info(), 1):
            if p_info.p_id == product_id:
                return num_products, num
        raise KeyError(product_id)

    def update(
        self,
        customisation_id: Optional[str],
        product_id: str,
        status: Optional[str] = None,
        download_states: Optional[Dict[str, str]] = None,
    ) -> None:
        with self._update_lock:
            with self.dict_from_file() as order:
                if status:
                    order["products_to_process"][product_id]["server_state"] = status
                if download_states:
                    order["products_to_process"][product_id]["download_states"] = download_states
                if customisation_id:
                    order["products_to_process"][product_id]["customisation"] = customisation_id
            self._locked_serialize(order)

    def reset_states(self) -> None:
        with self.dict_from_file() as order:
            products = order["products_to_process"]
        for prod_id, prod_info in products.items():
            if "download_states" in prod_info:
                del prod_info["download_states"]
            prod_info["server_state"] = "UNSUBMITTED"
            if order["type"] == "tailor":
                prod_info["customisation"] = None
        self._locked_serialize(order)

    def iter_product_info(self) -> Iterable[ProductInfo]:
        with self.dict_from_file() as order:
            orders = order["products_to_process"].items()
        for p_id, p_dict in orders:
            yield ProductInfo(p_id, p_dict)

    def get_products(self, datastore: Any) -> Iterable[Product]:
        for p_info in self.iter_product_info():
            yield datastore.get_product(p_info.p_dict["col_id"], p_info.p_id)

    def all_done(self) -> bool:
        (typ,) = self.get_dict_entries("type")
        if typ == "tailor":
            return self._all_done_tailor()
        elif typ == "download":
            return self._all_done_download()
        else:
            raise NotImplementedError(typ)

    def _all_done_tailor(self) -> bool:
        for p_info in self.iter_product_info():
            if not "download_states" in p_info.p_dict:
                return False
            for _fname, state in p_info.p_dict["download_states"].items():
                if state != "DOWNLOADED":
                    return False
        return True

    def _all_done_download(self) -> bool:
        return all([pi.p_dict["server_state"] == "DONE" for pi in self.iter_product_info()])


def _compute_state(p_dict: Dict[str, Any]) -> str:
    server_state = p_dict["server_state"]
    if server_state != "DONE":
        return server_state
    if not "download_states" in p_dict:
        return "DONE (NOT DOWNLOADED)"
    for _fname, state in p_dict["download_states"].items():
        if state != "DOWNLOADED":
            return "DONE (NOT DOWNLOADED)"
    return "DONE"


def get_default_order_dir() -> Path:
    order_dir = get_config_dir() / "orders"
    order_dir.mkdir(exist_ok=True, parents=True)
    return order_dir


def highest_number_in_order_filenames(file_paths: Iterable[Path]) -> int:
    number_pattern = re.compile(r".*#([\d]+).yml")
    order_numbers = [int(number_pattern.findall(fname.name)[0]) for fname in file_paths]
    return max(order_numbers)


def highest_prefix_in_order_filenames(file_paths: Iterable[Path]) -> str:
    fpaths = [f for f in file_paths if "#" in f.stem]
    return max([fpath.stem.split("#")[0] for fpath in fpaths])


def all_order_filenames(
    order_dir: Path,
    prefix: str = "",
) -> Iterable[Path]:
    glob = "*.yml"
    if len(prefix) > 0:
        glob = f"{prefix}#*.yml"
    return sorted(order_dir.glob(glob), key=_dt_from_order_filename)


def _dt_from_order_filename(fn: Path) -> datetime:
    try:
        date, number = fn.stem.split("#")
        return datetime.fromisoformat(date) + timedelta(milliseconds=int(number))
    except:
        return datetime.fromtimestamp(fn.stat().st_ctime)


def latest_order_file(order_dir: Path) -> Path:
    filepaths = all_order_filenames(order_dir)
    prefix = highest_prefix_in_order_filenames(filepaths)
    filepaths = all_order_filenames(order_dir, prefix)
    number = highest_number_in_order_filenames(filepaths)
    return order_dir / Path(f"{prefix}#{number:04d}.yml")


def new_order_filename(order_dir: Path) -> Path:
    with open_locked(order_dir / "lock") as lf:
        order_prefix = f"{datetime.now().strftime('%Y-%m-%d')}"

        all_filenames = list(all_order_filenames(order_dir, order_prefix))

        order_fn = order_dir / Path(f"{order_prefix}#{1:04d}.yml")
        if any(all_filenames):
            highest_existing_number = highest_number_in_order_filenames(all_filenames)

            order_fn = order_dir / Path(f"{order_prefix}#{int(highest_existing_number)+1:04d}.yml")
        order_fn.touch(exist_ok=False)
        return order_fn


def resolve_order(order_dir: Path, order_name: str) -> Order:
    if order_name == "latest":
        filenames = list(all_order_filenames(order_dir))
        if len(filenames) == 0:
            raise EumdacError("No order files found.")
        return Order(latest_order_file(order_dir))
    return Order(order_dir / Path(f"{order_name}.yml"))



from datetime import datetime

from eumdac.token import AccessToken
from eumdac.datastore import DataStore
from eumdac.collection import Collection, CollectionError
from eumdac.product import Product
from .base import INTEGRATION_TESTING, DataServiceTestCase


class TestCollection(DataServiceTestCase):
    collection_id = "EO:EUM:DAT:MSG:HRSEVIRI"

    def setUp(self):
        super().setUp()
        self.token = AccessToken(self.credentials)
        self.datastore = DataStore(token=self.token)
        self.collection = Collection(self.collection_id, self.datastore)

    def test_string_representation(self):
        self.assertEqual(self.collection_id, str(self.collection))
        self.assertIn(self.collection_id, repr(self.collection))

    def test_properties(self):
        self.assertIsInstance(self.collection.abstract, str)
        self.assertIsInstance(self.collection.title, str)
        self.assertIsInstance(self.collection.metadata, dict)
        self.assertIn("properties", self.collection.metadata)
        # check that the whitespaces have been fixed
        self.assertNotIn("\n", self.collection.abstract)

    def test_search(self):
        products = self.collection.search(
            dtstart=datetime(2022, 10, 1, 0, 0, 0), dtend=datetime(2022, 10, 1, 1, 0, 0)
        )
        product_ids = [str(product) for product in products]
        expected_product_ids = [
            "MSG4-SEVI-MSG15-0100-NA-20221001005743.037000000Z-NA",
            "MSG4-SEVI-MSG15-0100-NA-20221001004242.893000000Z-NA",
            "MSG4-SEVI-MSG15-0100-NA-20221001002742.762000000Z-NA",
            "MSG4-SEVI-MSG15-0100-NA-20221001001242.637000000Z-NA",
        ]
        self.assertEqual(product_ids, expected_product_ids)
        first = products.first()
        self.assertIsInstance(first, Product)

    def test_camparison(self):
        seviri_rss = Collection("EO:EUM:DAT:MSG:MSG15-RSS", self.datastore)
        avhrr_l1 = Collection("EO:EUM:DAT:METOP:AVHRRL1", self.datastore)
        self.assertLess(avhrr_l1, seviri_rss)
        self.assertNotEqual(avhrr_l1, seviri_rss)

    def test_product_type(self):
        self.assertEqual(self.collection.product_type, "HRSEVIRI")


class TestSearchResults(DataServiceTestCase):
    collection_id = "EO:EUM:DAT:MSG:HRSEVIRI"

    def setUp(self):
        super().setUp()
        self.token = AccessToken(self.credentials)
        self.datastore = DataStore(token=self.token)
        self.collection = Collection(self.collection_id, self.datastore)
        self.search_results = self.collection.search(
            dtstart=datetime(2022, 10, 1, 0, 0, 0), dtend=datetime(2022, 10, 1, 1, 0, 0)
        )

    def test_string_representation(self):
        self.assertIn(self.collection_id, repr(self.search_results))

    def test_properties(self):
        total_results = self.search_results.total_results
        all_results = list(self.search_results)
        self.assertEqual(total_results, len(all_results))

        query = self.search_results.query
        self.assertIn("dtstart", query)
        self.assertIn("dtend", query)

    def test_update_query(self):
        new_results = self.search_results.update_query(dtend=datetime(2022, 10, 1, 2, 0, 0))
        # check that we created a new instance
        self.assertIsNot(self.search_results, new_results)

    def test_first(self):
        first = self.search_results.first()
        all_products = list(self.search_results)
        self.assertEqual(first, all_products[0])
        self.assertIsInstance(first, Product)

    def test_empty_results(self):
        empty_results = self.collection.search(
            dtstart=datetime(2000, 1, 1, 0, 0, 0), dtend=datetime(2000, 1, 1, 0, 0, 1)
        )
        first = empty_results.first()
        self.assertIsNone(first)
        self.assertEqual(empty_results.total_results, 0)
        self.assertEqual(list(empty_results), [])

    def test_pagination(self):
        self.datastore.collections
        total_results = self.search_results.total_results

        self.search_results._items_per_page = total_results
        n_calls_before = len(self.requests_mock.calls)
        list(self.search_results)
        n_calls_after = len(self.requests_mock.calls)
        self.assertEqual(n_calls_after, n_calls_before + 1)

        self.search_results._items_per_page = total_results // 2
        n_calls_before = len(self.requests_mock.calls)
        list(self.search_results)
        n_calls_after = len(self.requests_mock.calls)
        self.assertEqual(n_calls_after, n_calls_before + 2)

    def test_results_contain_product(self):
        first = self.search_results.first()
        self.assertIn(first, self.search_results)

        product_id = "MSG1-SEVI-MSG15-0201-NA-20040120075736.727000000Z-NA"
        other = self.datastore.get_product(self.collection_id, product_id)
        self.assertNotIn(other, self.search_results)

    def test_invalid_search_argument(self):
        with self.assertRaisesRegex(CollectionError, r"invalid search options .*"):
            self.collection.search(foo="bar")

    def test_search_no_starttime(self):
        results = self.collection.search(dtend=datetime(2022, 10, 1, 1, 0, 0))

    def test_search_no_endtime(self):
        results = self.collection.search(dtstart=datetime(2000, 1, 1, 0, 0, 0))



"""
Module: eumdac.futures

This module defines classes for managing a custom thread pool executor with cooperative function handling.

Classes:
- EumdacFutureFunc: Represents a callable function with cooperative handling.
- EumdacThreadPoolExecutor: Extends ThreadPoolExecutor to manage cooperative function execution.

Usage:
1. Create instances of EumdacFutureFunc to define callable functions with cooperative handling.
2. Use EumdacThreadPoolExecutor to submit functions to a thread pool with cooperative handling.
"""

import concurrent.futures
import sys
from typing import Any, List
from eumdac.logging import logger


class EumdacFutureFunc:
    def __init__(self) -> None:
        """
        Initialize the EumdacFutureFunc object.
        """
        self.aborted = False

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """
        Placeholder for the callable function. Must be implemented in subclasses.

        Raises:
        - NotImplementedError: If the method is called directly without being implemented in a subclass.
        """
        raise NotImplementedError()

    def abort(self) -> None:
        """
        Set the 'aborted' flag to True, indicating that the function has been aborted.
        This needs to be handled cooperatively in the Subclasses
        """
        logger.debug(f"{self} abort request received")
        self.aborted = True


class EumdacThreadPoolExecutor(concurrent.futures.ThreadPoolExecutor):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Initialize the EumdacThreadPoolExecutor object.

        Attributes:
        - functors (List[EumdacFutureFunc]): List to store EumdacFutureFunc instances.
        """
        self.functors: List[EumdacFutureFunc] = []
        super().__init__(*args, **kwargs)

    def pool_shutdown(self) -> None:
        """
        Abort all functions in the 'functors' list and ask them to gracefully shut down.
        """
        logger.debug(f"{self} pool_shutdown issued")
        for f in self.functors:
            logger.debug(f"{self} aborting {f}")
            f.abort()
        if sys.version_info >= (3, 9):
            return super().shutdown(wait=True, cancel_futures=True)
        else:
            return super().shutdown(wait=True)

    def pool_submit(
        self, fn: EumdacFutureFunc, *args: Any, **kwargs: Any
    ) -> "concurrent.futures.Future[Any]":
        """
        Submit a function to the thread pool executor and add it to the 'functors' list.

        Args:
        - fn (EumdacFutureFunc): The function to be submitted.
        - *args: Variable length argument list.
        - **kwargs: Arbitrary keyword arguments.

        Returns:
        - concurrent.futures.Future[Any]: A Future object representing the execution of the submitted function.
        """
        self.functors.append(fn)
        return super().submit(fn, *args, **kwargs)



"""Module containing package information for setup.py and __init__.py"""

__title__ = "eumdac"
__description__ = "EUMETSAT Data Access Client"
__url__ = "https://gitlab.eumetsat.int/eumetlab/data-services/eumdac"
__version__ = "3.0.0"
__documentation__ = (
    "https://user.eumetsat.int/resources/user-guides/eumetsat-data-access-client-eumdac-guide"
)
__api_documentation__ = "https://usc.tools.eumetsat.int/docs/eumdac/"
__author__ = "EUMETSAT"
__author_email__ = "ops@eumetsat.int"
__license__ = "MIT"



"""module containing the TailorApp which will be used when using 
eumdac download **with** the --tailor argument."""

from __future__ import annotations

import concurrent
import fnmatch
import re
import shutil
import tempfile
import threading
import time
import typing
from collections import namedtuple
from datetime import timedelta
from pathlib import Path
from typing import Any, Callable, Dict, Generator, Iterable, List, Optional, Tuple

from eumdac.config import PERCENTAGE_WARNING
from eumdac.customisation import (
    Customisation,
    CustomisationError,
    UnableToGetCustomisationError,
)
from eumdac.datastore import DataStore
from eumdac.datatailor import DataTailor
from eumdac.errors import EumdacError
from eumdac.futures import EumdacFutureFunc, EumdacThreadPoolExecutor
from eumdac.job_id import JobIdentifier
from eumdac.logging import logger
from eumdac.order import Order
from eumdac.product import Product
from eumdac.tailor_models import Chain


class FailedStateTransitionError(Exception):
    def __init__(self, msg: str, faillog: str) -> None:
        self.faillog = faillog
        super().__init__(msg)

    def find_lines(self, search_string: str) -> Generator[str, None, None]:
        if self.faillog:
            for line in self.faillog.splitlines():
                if search_string in line:
                    result_text = line.split(" - ")[-1]
                    yield result_text


class TailorApp:
    def __init__(
        self,
        order: Order,
        datastore: Any,
        datatailor: Any,
    ) -> None:
        self.server_state_monitor_executor = EumdacThreadPoolExecutor(max_workers=3)
        self.download_executor = EumdacThreadPoolExecutor(max_workers=2)
        self.datastore = datastore
        self.datatailor = datatailor
        self.order = order

        with self.order.dict_from_file() as order_d:
            self.output_dir = Path(order_d["output_dir"])
            self.output_dir.mkdir(exist_ok=True, parents=True)
            self.chain = Chain(**order_d["chain"])

    def run(self) -> bool:
        try:
            return self.resume()
        except FatalEumdacError as fee:
            logger.error(f"Fatal error during execution: {fee}")
            self.order.remote_delete_failed(self.datatailor)

        return False

    def shutdown(self) -> None:
        self.server_state_monitor_executor.pool_shutdown()
        self.download_executor.pool_shutdown()
        with self.order._lock:
            return

    def concurrent_download(
        self,
        products: Iterable[Product],
        output_dir: Path,
        customization_add_func: Callable[[Product], Customisation],
        timeout: float = 600,
    ) -> None:
        customisation_futures: List[concurrent.futures.Future[Any]] = []
        download_futures = []

        failed_customisations = []
        done_customistations = []

        num_jobs = len(list(self.order.iter_product_info()))
        job_identificator = JobIdentifier(num_jobs)

        for product in products:
            customisation_futures.append(
                self.server_state_monitor_executor.pool_submit(
                    WaitForDoneCustomisationFutureFunc(
                        customization_add_func,
                        self.order,
                        job_identificator,
                    ),
                    product,
                )
            )

        while customisation_futures:
            done_concurrent_futures: List[concurrent.futures.Future[Any]] = []
            not_done_concurrent_futures: List[
                concurrent.futures.Future[Any]
            ] = customisation_futures
            try:
                (
                    done_concurrent_futures,
                    not_done_concurrent_futures,
                ) = [
                    list(x)
                    for x in concurrent.futures.wait(
                        customisation_futures,
                        return_when=concurrent.futures.FIRST_COMPLETED,
                        timeout=None,
                    )
                ]
                for future in done_concurrent_futures:
                    customisation_futures.remove(future)
            except concurrent.futures.TimeoutError:
                pass

            # at this point done_concurrent_futures contain all finished customisations
            # now check if they failed and submit a task to download the result if success
            for done_future in done_concurrent_futures:
                try:
                    completed_customistation, product = done_future.result()
                    done_customistations.append(completed_customistation)
                except GracefulAbortError as graceful_abort_error:
                    logger.debug(f"External abort for {done_future}: {graceful_abort_error}")
                    continue
                except FatalEumdacError as fatal_eumdac_error:
                    for future in customisation_futures:
                        future.cancel()
                    self.server_state_monitor_executor.pool_shutdown()
                    if fatal_eumdac_error.extra_info:
                        logger.error(
                            f"Fatal error: {fatal_eumdac_error} - {fatal_eumdac_error.extra_info['title']}: {fatal_eumdac_error.extra_info['description']}"
                        )
                    else:
                        logger.error(f"Fatal error: {fatal_eumdac_error}")
                    raise
                except CustomisationTimeoutError as te:
                    logger.error(f"{completed_customistation} timed out: {te}")
                    failed_customisations.append(completed_customistation)
                    continue
                except CustomisationError as ce:
                    logger.error(f"Failed: {ce}")
                    continue
                except Exception as exc:
                    logger.error(f"{done_future} failed: {exc}")
                    failed_customisations.append(done_future)
                    continue

                logger.debug(f"{completed_customistation} processed and ready to download.")
                download_futures.append(
                    self.download_executor.pool_submit(
                        DownloadRunFutureFunc(
                            self.order,
                            output_dir,
                            job_identificator.job_id_str(product),
                        ),
                        completed_customistation,
                        product,
                    )
                )

        _ = concurrent.futures.wait(download_futures, return_when=concurrent.futures.ALL_COMPLETED)

    def resume(self) -> bool:
        # handle existing customisations
        customisations = []
        products_to_resume = []
        products_to_repeat = []

        # query all customisation states
        success_customisation_product_futures = []

        try:
            user_name = self.datatailor.user_info["username"]
            quota_info = self.datatailor.quota["data"][user_name]
            if quota_info["space_usage_percentage"] > PERCENTAGE_WARNING:
                logger.warning(f"Reaching maximum quota: {quota_info['space_usage_percentage']}%")
            elif quota_info["space_usage_percentage"] > 100:
                logger.warning(f"Over maximum quota: {quota_info['space_usage_percentage']}%")
        except EumdacError as e:
            # The quota call is unsupported by local-tailor, so we don't report
            if not self.datatailor.is_local:
                logger.warning(f"Could not determine current quota: {e}")

        for p_info in self.order.iter_product_info():
            success_customisation_product_futures.append(
                self.server_state_monitor_executor.pool_submit(
                    StateQueryFutureFunc(),
                    p_info.p_id,
                    p_info.p_dict,
                    self.datatailor,
                    self.datastore,
                )
            )

        done_success_customisation_product_futures, _ = concurrent.futures.wait(
            success_customisation_product_futures,
            return_when=concurrent.futures.ALL_COMPLETED,
        )
        for done_success_customisation_product_future in done_success_customisation_product_futures:
            (
                success,
                customisation,
                product,
            ) = done_success_customisation_product_future.result()
            if success:
                customisations.append(customisation)
                products_to_resume.append(product)
            else:
                products_to_repeat.append(product)

        if len(products_to_resume) > 0:
            self.concurrent_download(
                products_to_resume,
                self.output_dir,
                customization_add_func=GetCustomisation(customisations),
            )
        if len(products_to_repeat) > 0:
            self.concurrent_download(
                products_to_repeat,
                self.output_dir,
                customization_add_func=lambda x: self.datatailor.new_customisation(x, self.chain),
            )

        return True


class DownloadRunFutureFunc(EumdacFutureFunc):
    def __init__(self, order: Order, output_dir: Path, job_id: str):
        super().__init__()
        self.order = order
        self.output_dir = output_dir
        self.job_id = job_id

    @typing.no_type_check
    def __call__(
        self,
        customisation: Customisation,
        product: Product,
    ) -> None:
        results = customisation.outputs
        # compare with file_patterns from order_file
        (file_patterns,) = self.order.get_dict_entries("file_patterns")
        if file_patterns:
            filtered_results = []
            for pattern in file_patterns:
                matches = fnmatch.filter(results, pattern)
                filtered_results.extend(matches)
            results = filtered_results

        logger.debug(f"{self.job_id}: Starting download(s) for {results}")
        download_states = {result: "PENDING" for result in results}
        self.order.update(
            customisation._id,
            product._id,
            status=None,
            download_states=download_states,
        )

        with self.order.dict_from_file() as order_d:
            dirs = order_d["dirs"]

        num_total, num = self.order.resolve_product_num(product._id)
        for result in results:
            if self.aborted:
                break
            download_states[result] = "DOWNLOAD_ERROR"
            try:
                logger.info(
                    f"{self.job_id}: Downloading output of job {customisation._id} for {product._id}"
                )
                self.download_customisation_result(customisation, result, dirs)
                logger.info(f"{self.job_id}: {Path(result).parts[-1]} has been downloaded.")
                download_states[result] = "DOWNLOADED"
            except DownloadExistsError as err:
                logger.info(f"{self.job_id}: Skipping download. File exists: {err.product_path}")
                download_states[result] = "DOWNLOADED"
            except DownloadAbortedError as err:
                logger.warning(f"{self.job_id}: Download of {err.product_path} aborted")
            except Exception as exc:
                logger.error(f"{self.job_id}: Error while downloading: {exc}")

            self.order.update(
                customisation._id,
                product._id,
                status=None,
                download_states=download_states,
            )

        # delete serverside customisation on success
        if "DOWNLOAD_ERROR" not in download_states.values():
            logger.info(
                f"{self.job_id}: Deleting customization {customisation._id} for {product._id}"
            )
            customisation.delete()
        else:
            logger.warning(
                f"{self.job_id}: {customisation} download failed. Keeping customisation."
            )

    fatal_error_logs = {
        "ERROR": ["invalid", "INTERNAL ERROR"],
    }

    def download_customisation_result(
        self, customisation: Customisation, result: str, dirs: bool
    ) -> None:
        product_path = self.output_dir / Path(result).parts[-1]

        if dirs:
            # when the dirs flag is used
            # a subdirectory is created
            # to avoid overwriting common files
            output_subdir = self.output_dir / f"{Path(result).parts[0]}"
            output_subdir.mkdir(exist_ok=True)
            product_path = output_subdir / Path(result).parts[-1]

        if product_path.is_file():
            raise DownloadExistsError(product_path)

        with tempfile.TemporaryDirectory(dir=self.output_dir, suffix=".tmp") as tempdir:
            tmp_prod_p = Path(tempdir) / product_path.parts[-1]
            with tmp_prod_p.open("wb") as tmp_prod:
                with customisation.stream_output_iter_content(result) as chunks:
                    for chunk in chunks:
                        if self.aborted:
                            raise DownloadAbortedError(product_path)
                        tmp_prod.write(chunk)
            shutil.move(str(tmp_prod_p), str(product_path))


def check_invalid_state_transition_changelog(error: FailedStateTransitionError) -> None:
    fatal_error_logs = {
        "ERROR": ["INTERNAL ERROR", "incompatible"],
    }
    # find messages containing errors
    for severity, filters in fatal_error_logs.items():
        for line in error.find_lines(severity):
            if any(f in line for f in filters):
                desc = {"status": 200, "title": severity, "description": line}
                raise FatalEumdacError(EumdacError(line, desc))


class StateQueryFutureFunc(EumdacFutureFunc):
    @typing.no_type_check
    def __call__(
        self,
        p_id: str,
        p_info: Dict[str, Any],
        datatailor: DataTailor,
        datastore: DataStore,
    ) -> Tuple[bool, Optional[Customisation], Optional[Product]]:
        if self.aborted:
            return False, None, None
        product = datastore.get_product(p_info["col_id"], p_id)
        customisation_id = p_info["customisation"]
        if customisation_id is None:
            return False, None, product
        customisation = datatailor.get_customisation(customisation_id)
        success = False
        try:
            _ = customisation.status
            success = True
        except (CustomisationError, CustomisationTimeoutError) as ce:
            download_states = {}
            if "download_states" in p_info:
                download_states = p_info["download_states"]

            if "DOWNLOADED" in download_states.values():
                logger.warning(
                    f"Customisation {customisation_id} has already been finished and downloaded."
                )
            else:
                logger.warning(f"Could not restore customisation for {customisation_id}: {ce}")
                try:
                    if customisation:
                        customisation.delete()
                except:
                    pass

        return success, customisation, product


class GetCustomisation:
    def __init__(self, customisations: List[Customisation]):
        self.cnt = 0
        self.customisations = customisations
        self.lock = threading.Lock()

    def __call__(self, _product: Product) -> Customisation:
        with self.lock:
            result = self.customisations[self.cnt]
            if self.cnt < len(self.customisations):
                self.cnt += 1
            return result


State = namedtuple("State", "name log")


class WaitForDoneCustomisationFutureFunc(EumdacFutureFunc):
    def __init__(
        self,
        customization_add_func: Callable[[Product], Customisation],
        order: Order,
        job_identificator: JobIdentifier,
        timeout: float = 1800,
        polling_interval: float = 1.0,
        max_retries: int = 10,
        max_timeouts: int = 3,
        wait_timedelta: timedelta = timedelta(seconds=5.0),
    ) -> None:
        super().__init__()
        self.customization_add_func = customization_add_func
        self.order = order
        self.job_identificator = job_identificator
        self.timeout = timeout
        self.polling_interval = polling_interval
        self.max_retries = max_retries
        self.max_timeouts = max_timeouts
        self.wait_timedelta = wait_timedelta

        self.terminated = False
        self.failed = False
        self.state = State("UNSUBMITTED", "")
        self.state_transitions: Dict[str, List[str]] = {
            "UNSUBMITTED": ["QUEUED", "RUNNING", "DONE", "FAILED", "INACTIVE"],
            "QUEUED": ["QUEUED", "RUNNING", "DONE", "FAILED", "INACTIVE"],
            "RUNNING": ["RUNNING", "DONE", "FAILED", "INACTIVE"],
            "DONE": [],
            "FAILED": [],
            "INACTIVE": [],
        }
        self.timed_out = False
        self.job_identificator = job_identificator
        self.job_id = "(? / ?)"

    @typing.no_type_check
    def __call__(
        self,
        product: Product,
    ) -> Optional[Customisation]:
        if self.aborted:
            return

        # Set the job id only when the Future is actually called by the Scheduler to ensure
        # (somewhat) correct ordering.
        self.job_identificator.register(product)
        self.job_id = self.job_identificator.job_id_str(product)

        logger.info(f"Triggering {self.job_id.lower()} of {self.job_identificator.total_jobs}")
        logger.debug(
            f"{self.job_id} tracked as {self.job_identificator.registered_objects[product]}"
        )
        customisation = self.try_to_add_customisation(product)
        return self.wait_for_success(customisation, product)

    def try_to_add_customisation(self, product: Product) -> Optional[Customisation]:
        retries = 0
        while True:
            if self.aborted:
                raise GracefulAbortError("Abort requested.")

            if retries >= self.max_retries:
                raise CustomisationError(
                    f"{self.job_id}: Could not add customisation after {retries} retries"
                )
            try:
                return self.customization_add_func(product)
            except EumdacError as e:
                retries += 1
                try:
                    check_error_response(e)
                except ExceedingNumberOfCustomisationsEumdacError as _exceeding_number_error:
                    pass

                logger.warning(
                    f"{self.job_id}: {e}: Could not create customisation. Retry: {retries}"
                )
                time.sleep(self.wait_timedelta.total_seconds())

    def wait_for_success(
        self, customisation: Customisation, product: Product
    ) -> Tuple[Customisation, Product]:
        retries = 0
        timeouts = 0
        while retries < self.max_retries:
            if self.aborted:
                break
            self.state_timer = self.windup_timer()
            try:
                while not self.aborted and not self.terminated:
                    self.step(customisation, product._id)
                    if self.timed_out:
                        logger.error(f"{self.job_id}: {self} timed_out")
                        retries += 1
                        timeouts += 1
                        if timeouts >= self.max_timeouts:
                            raise FatalEumdacError(
                                EumdacError(f"{self.job_id}: {customisation._id} timed out")
                            )
                break
            except FailedStateTransitionError as fste:
                logger.debug(f"{fste}: {fste.faillog}")
                check_invalid_state_transition_changelog(fste)

                if customisation:
                    if self.state.name == "INACTIVE":
                        customisation.kill()
                    customisation.delete()

                retries += 1
                logger.debug(
                    f"{self.job_id}: {self} failed with {fste} on try {retries}/{self.max_retries}"
                )
                if retries < self.max_retries:
                    # if we retry this reinitialize state
                    self.state = State("UNSUBMITTED", "")

                while retries < self.max_retries:
                    ret_customisation = self.try_to_add_customisation(product)
                    if ret_customisation:
                        customisation = ret_customisation
                        break
                    retries += 1

            finally:
                self.state_timer.cancel()
                if retries >= self.max_retries:
                    raise FatalEumdacError(
                        EumdacError(f"{self.job_id}: {customisation._id} is inactive")
                    )

        return customisation, product

    def step(self, customisation: Customisation, product_id: str) -> None:
        self.transition(customisation, product_id)

        self.terminated = self.state.name == "DONE"
        self.failed = self.state.name in ["FAILED", "INACTIVE"]
        if self.terminated:
            return
        if self.failed:
            logs = self.get_latest_log(customisation)
            if not self.order._no_warning_logs:
                log_message = logs.split("\n\n")
                logger.warning(f"{self.job_id}: {log_message[-1]}")
            raise FailedStateTransitionError(
                f"{self.job_id}: Server state is: {self.state.name}",
                faillog=logs,
            )

        if self.state.name not in self.state_transitions.keys():
            raise InvalidStateTransitionError(
                "{self.job_id}: Unexpected State: {self.state.name})",
                old_state=self.state,
                new_state=State(self.state.name, self.get_latest_log(customisation)),
            )

    def transition(
        self,
        customisation: Customisation,
        product_id: str,
    ) -> None:
        new_state = customisation.status
        if self.state.name != new_state:
            logger.debug(
                f"{self.job_id}: {customisation}: State change: {self.state.name} -> {new_state}"
            )
            num_total, num = self.order.resolve_product_num(product_id)
            suffix = {
                "QUEUED": "is now queued",
                "RUNNING": "is now running",
                "DONE": "has finished",
                "FAILED": "has failed",
                "INACTIVE": "is inactive",
            }
            logger.info(
                f"{self.job_id}: Customisation {customisation._id} for {product_id} {suffix[new_state]}"
            )

        if new_state not in self.state_transitions[self.state.name]:
            self.state_timer.cancel()
            raise InvalidStateTransitionError(
                f"{self.job_id}: Tried to transition from state {self.state} to {new_state}, "
                "which is not expected",
                old_state=self.state,
                new_state=State(new_state, self.get_latest_log(customisation)),
            )
        elif self.state.name != new_state:
            self.state_timer.cancel()
            self.state_timer = self.windup_timer()
            self.state = State(new_state, self.get_latest_log(customisation))
            if self.order:
                self.order.update(
                    customisation._id,
                    product_id,
                    self.state.name,
                )

    def windup_timer(self) -> threading.Timer:
        logger.debug(f"{self.job_id}: {repr(self)} windup_timer")
        timer = threading.Timer(self.timeout, self.on_timeout)
        timer.start()
        return timer

    def on_timeout(self) -> None:
        logger.debug(f"{self.job_id}: {repr(self)} on_timeout")
        self.timed_out = True

    def get_latest_log(self, customisation: Customisation) -> str:
        try:
            return customisation.logfile
        except UnableToGetCustomisationError:
            return "Unable to get logs"


class InvalidStateTransitionError(EumdacError):
    def __init__(self, msg: str, old_state: State, new_state: State) -> None:
        self.old_state = old_state
        self.new_state = new_state
        super().__init__(msg)


class CustomisationTimeoutError(EumdacError):
    """Error raised during downloads"""


class DownloadError(EumdacError):
    """Error raised during downloads"""

    def __init__(self, product_path: Path, *args: Any) -> None:
        super().__init__(msg=str(product_path))
        self.product_path = product_path


class DownloadAbortedError(DownloadError):
    """Error raised when a download is aborted"""


class DownloadExistsError(DownloadError):
    """Error raised when a download file already exists"""


class GracefulAbortError(EumdacError):
    """Error related to abort conditions when creating customisations"""


class FatalEumdacError(EumdacError):
    """Unrecoverable Error"""

    def __init__(self, eumdac_error: EumdacError):
        super().__init__(eumdac_error.args[0], eumdac_error.extra_info)


class ExceedingNumberOfCustomisationsEumdacError(EumdacError):
    """Error for server responding that maximum number of customisations is reached"""

    def __init__(self, eumdac_error: EumdacError, number_of_customisations: int):
        super().__init__(eumdac_error.args[0], eumdac_error.extra_info)
        self.number_of_customisations = number_of_customisations


def check_error_response(error: Exception) -> None:
    """helper function to check error responses from the server"""
    fatal_error_responses = {
        400: [""],
    }

    number_of_customisation_responses = {
        500: ["You are exceeding your maximum number"],
    }

    if isinstance(error, EumdacError):
        if (
            error.extra_info
            and error.extra_info["status"] in fatal_error_responses
            and "description" in error.extra_info
        ):
            if any(
                fatal_response in error.extra_info["description"]
                for fatal_response in fatal_error_responses[error.extra_info["status"]]
            ):
                raise FatalEumdacError(error)

        if (
            error.extra_info
            and error.extra_info["status"] in number_of_customisation_responses
            and "description" in error.extra_info
        ):
            if any(
                number_response in error.extra_info["description"]
                for number_response in number_of_customisation_responses[error.extra_info["status"]]
            ):
                response_match = re.search(r"([\d]+)", error.extra_info["description"])
                if response_match:
                    number = int(response_match[0])
                else:
                    number = -1
                raise ExceedingNumberOfCustomisationsEumdacError(error, number)

    else:
        return



# type: ignore
"""Fake DataStore, DataTailor and Product that will be used when adding the --test
option in favour of the real implementations. Only useful for unittests."""

import io
from contextlib import contextmanager


class FakeDataStore:
    """Fake DataStore for testing."""

    def get_collection(self, collection_id):
        """Return a FakeCollection with `collection_id`."""
        return FakeCollection(collection_id)

    def get_product(self, collection_id, product_id):
        """Return a FakeProduct with `product_id` from `collection_id`."""
        return FakeProduct(collection_id, product_id)


class FakeProduct:
    """FakeProduct for testing."""

    def __init__(self, collection_id, product_id):
        """Init from `collection_id` and `product_id`."""
        self._id = product_id
        self.collection = FakeCollection(collection_id)
        self.entries = ["entry1.nc", "entry2.nc"]

    def __str__(self):
        """Return the id as str representation"""
        return str(self._id)

    def open(self, entry=None, chunk=None, custom_headers=None):
        """Return a fake stream as the contents of the product."""
        if entry:
            return FakeStream(f"{self._id}-{entry}")

        return FakeStream(self._id)

    @property
    def md5(self):
        """Return the md5 of the fake stream returned on open."""
        import hashlib

        with self.open(None) as f:
            return hashlib.md5(f.read()).hexdigest()


class FakeStream:
    def __init__(self, name):
        self.decode_content = True
        self.name = name
        self.content = io.BytesIO(b"Content")

    def getheader(self, header):
        if header == "Content-Length":
            # Return a fixed length (7) for 'Content-Length' header.
            return 7
        return None

    def read(self, num=None):
        return self.content.read(num)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class FakeCollection:
    """Fake Collection for testing."""

    def __init__(self, collection_id):
        """Init from `collection_id`."""
        self._id = collection_id

    def __str__(self):
        """Return id as the str representation."""
        return str(self._id)

    def search(self, **query):
        """Return fake search results."""
        dtstart = query["dtstart"]
        dtend = query["dtend"]
        return [
            FakeProduct(self._id, f"prod_{dtstart.isoformat().strip().replace(':', '-')}"),
            FakeProduct(self._id, f"prod_{dtend.isoformat().strip().replace(':', '-')}"),
        ]


class FakeDataTailor:
    """Fake DataTailor for testing."""

    pass



"""Module for interfacing with local Data Tailor instances."""

from pathlib import Path
from urllib.parse import urlparse

from eumdac.config import get_url_path
from eumdac.datatailor import DataTailor
from eumdac.errors import EumdacError
from eumdac.token import AnonymousAccessToken, URLs

import sys

if sys.version_info < (3, 9):
    from typing import Iterable
else:
    from collections.abc import Iterable


def get_tailor_id(filepath: Path) -> str:
    """get a tailor id from a configuration file path"""
    return filepath.stem


def get_tailor_path(tailor_id: str) -> Path:
    """get a configuration file path from a tailor id"""
    for fn in all_url_filenames():
        if tailor_id == get_tailor_id(fn):
            return fn

    raise EumdacError(f"local-tailor id not found: {tailor_id}")


def get_urls(filepath: Path) -> URLs:
    """retrieve a URLs instance from an INI file path"""
    return URLs(str(filepath))


def all_url_filenames(prefix: str = "") -> Iterable[Path]:
    """retrieve file Paths of configured URL files"""
    url_dir = get_url_path()
    glob = "*.ini"
    if len(prefix) > 0:
        glob = f"{prefix}#*.ini"
    return sorted(url_dir.glob(glob))


def new_url_filename(tailor_id: str) -> Path:
    """create a Path pointing to a URLs configuration INI file that can be created subsequently"""
    return get_url_path() / Path(f"{tailor_id}.ini")


def remove_url(url_name: str) -> None:
    """remove a URLs configuration INI from configuration directory"""
    p = get_url_path() / Path(f"{url_name}.ini")
    if p.exists():
        p.unlink()


def resolve_url(url_name: str) -> URLs:
    """retrieve a URLs instance from an identifier"""
    p = get_url_path() / Path(f"{url_name}.ini")
    if p.exists():
        return URLs(str(p))
    else:
        raise EumdacError(f"{url_name} not found at {str(p)}.")


def get_local_tailor(tailor_id: str) -> DataTailor:
    """create a DataTailor instance using a URLs configuration identified via configuration identifier"""
    url = resolve_url(tailor_id)
    token = AnonymousAccessToken(urls=url)
    return DataTailor(token)


def new_local_tailor(tailor_id: str, tailor_url: str) -> Path:
    """create a configuration for a local-tailor instance specifying an identifer and base url"""
    parsed_url = urlparse(tailor_url)

    if not parsed_url.scheme:
        raise EumdacError("No scheme provided")
    if not parsed_url.hostname:
        raise EumdacError("No hostname provided")
    if not parsed_url.port:
        raise EumdacError("No port provided")

    filepath = new_url_filename(tailor_id)
    with filepath.open("w") as f:
        new_url = URLs()
        new_url.set("tailor", "epcs", f"{tailor_url}/api/v1")
        new_url.write(f)
    return filepath


def remove_local_tailor(tailor_id: str) -> None:
    """remove a local-tailor configuration by specifying its' identifier"""
    remove_url(tailor_id)


def is_online(filepath: Path) -> bool:
    """determine the state of a local-tailor instance by specifying Path to configuration file"""
    try:
        dt = get_local_tailor(get_tailor_id(filepath))
        _ = dt.info
        return True
    except Exception:
        return False


def get_api_url(filepath: Path) -> str:
    """retrieve local-tailor base api url from a given local-tailor configuration file path"""
    urls = get_urls(filepath)
    return urls.get("tailor", "epcs")



from datetime import datetime

from eumdac.token import AccessToken
from eumdac.datastore import DataStore
from eumdac.product import Product
from .base import DataServiceTestCase


class TestProduct(DataServiceTestCase):
    collection_id = "EO:EUM:DAT:MSG:HRSEVIRI"
    product_id = "MSG4-SEVI-MSG15-0100-NA-20221010104242.601000000Z-NA"
    product_manifest = b"""<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<informationPackage xmlns="http://www.eumetsat.int/sip">
    <dataSection>
        <dataObject ID="MSG4-SEVI-MSG15-0100-NA-20221010104242.601000000Z-NA.nat">
            <fileType>Data</fileType>
            <format>Native</format>
            <size>271175723</size>
            <path>MSG4-SEVI-MSG15-0100-NA-20221010104242.601000000Z-NA.nat</path>
        </dataObject>
    </dataSection>
    <eopmetadataSection>
        <path>EOPMetadata.xml</path>
    </eopmetadataSection>
</informationPackage>
"""

    def setUp(self):
        super().setUp()
        self.token = AccessToken(self.credentials)
        self.datastore = DataStore(token=self.token)
        self.product = Product(self.collection_id, self.product_id, self.datastore)

    def test_string_representation(self):
        self.assertEqual(self.product_id, str(self.product))
        self.assertTrue(self.product_id in repr(self.product))

    def test_download_entry(self):
        with self.product.open("manifest.xml") as file:
            filename = file.name
            content = file.read()
        self.assertEqual(filename, "manifest.xml")
        self.assertEqual(content, self.product_manifest)

    def test_properties(self):
        collection = self.product.collection
        self.assertEqual(str(collection), self.collection_id)
        self.assertIsInstance(self.product.sensing_start, datetime)
        self.assertIsInstance(self.product.sensing_end, datetime)
        self.assertTrue(self.product.satellite.startswith("MSG"))
        self.assertEqual(self.product.instrument, "SEVIRI")
        self.assertIsInstance(self.product.size, int)
        self.assertIsInstance(self.product.metadata, dict)
        self.assertIn("properties", self.product.metadata)
        self.assertIn("manifest.xml", self.product.entries)
        self.assertIsInstance(self.product.acronym, str)
        self.assertIsInstance(self.product.product_type, str)
        self.assertEqual(self.product.product_type, self.product.acronym)
        self.assertIsNone(self.product.timeliness)
        self.assertIsInstance(self.product.md5, str)
        self.assertIsInstance(self.product.processingTime, str)
        self.assertIsInstance(self.product.processorVersion, str)
        self.assertIsInstance(self.product.format, str)
        self.assertIsInstance(self.product.qualityStatus, str)
        self.assertIsInstance(self.product.ingested, datetime)
        self.assertIsInstance(self.product.orbit_type, str)
        self.assertIsInstance(self.product.orbit_is_LEO, bool)
        self.assertIsInstance(self.product.url, str)
        # These are none for GEO products
        self.assertIsNone(self.product.orbit_number)
        self.assertIsNone(self.product.orbit_direction)
        self.assertIsNone(self.product.relative_orbit)
        self.assertIsNone(self.product.cycle_number)

    def test_camparison(self):
        seviri_rss = Product(
            "EO:EUM:DAT:MSG:MSG15-RSS",
            "MSG2-SEVI-MSG15-0100-NA-20211013140918.098000000Z-NA",
            self.datastore,
        )
        avhrr_l1 = Product(
            "EO:EUM:DAT:METOP:AVHRRL1",
            "AVHR_xxx_1B_M01_20211014082503Z_20211014100403Z_N_O_20211014091048Z",
            self.datastore,
        )
        avhrr_l1_other = Product(
            "EO:EUM:DAT:METOP:AVHRRL1",
            "AVHR_xxx_1B_M01_20210206061303Z_20210206075503Z_N_O_20210206065931Z",
            self.datastore,
        )
        self.assertLess(avhrr_l1, seviri_rss)
        self.assertNotEqual(avhrr_l1, seviri_rss)
        self.assertLess(avhrr_l1_other, avhrr_l1)
        self.assertNotEqual(avhrr_l1_other, avhrr_l1)



import unittest
import re
import pickle
import os
import io
import gzip
import zlib
import urllib
from base64 import b64encode

import responses
from responses import matchers
from contextlib import contextmanager


CONSUMER_KEY = os.getenv("CONSUMER_KEY")
CONSUMER_SECRET = os.getenv("CONSUMER_SECRET")

# different modes
# INTEGRATION_TESTING -> Targets real OPE endpoints + record all calls
INTEGRATION_TESTING = os.getenv("INTEGRATION_TESTING")
if INTEGRATION_TESTING and not (CONSUMER_KEY and CONSUMER_SECRET):
    raise RuntimeError("Integration testing requires credentials!")

http_methods = ["GET", "HEAD", "POST", "PUT", "DELETE", "CONNECT", "OPTIONS", "TRACE", "PATCH"]


class ReplayResponse(responses.Response):
    def __init__(self, request, response):
        self._request = request
        self._response = response
        parsed_url = urllib.parse.urlparse(request.url)
        super().__init__(
            request.method,
            request.url,
            body=response.content,
            status=response.status_code,
            headers=response.headers,
            match=[matchers.query_string_matcher(parsed_url.query)],
        )
        if "gzip" in response.headers.get("Content-Encoding", ""):
            self.body = self._gzip_compress(self.body)

    @staticmethod
    def _gzip_compress(data):
        if hasattr(data, "read"):
            data = data.read()
        if isinstance(data, str):
            data = data.encode()
        # window size flag, +25 to +31 include a basic gzip
        # header and trailing checksum
        wbits = 28
        compressor = zlib.compressobj(wbits=wbits)
        compressed_chunks = [compressor.compress(data), compressor.flush()]
        return b"".join(compressed_chunks)


class ReplayResponsesTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if INTEGRATION_TESTING:
            cls.prepare_integration_test()
        cls.recordpath = "{0}/data/{1}.{2}.pickle.gz".format(
            os.path.dirname(__file__), cls.__module__.split(".")[-1], cls.__name__
        )
        cls.requests_mock = responses.RequestsMock()
        if INTEGRATION_TESTING:
            cls._records = {}
        else:
            with gzip.open(cls.recordpath) as file:
                cls._records = pickle.load(file)

    def setUp(self):
        if INTEGRATION_TESTING:
            url = re.compile(".*")
            for method in http_methods:
                self.requests_mock.add(responses.PassthroughResponse(method, url))
            self.requests_mock.assert_all_requests_are_fired = False
            self.requests_mock.response_callback = self._record_streaming_response_body
        else:
            calls = self._records.get(self.record_key, responses.CallList())
            for request, response in calls:
                replay = ReplayResponse(request, response)
                self.requests_mock.add(replay)
            self.requests_mock.assert_all_requests_are_fired = True
        self.addCleanup(self.requests_mock.reset)
        self.requests_mock.start()
        self.addCleanup(self.requests_mock.stop)

    def tearDown(self):
        if INTEGRATION_TESTING:
            calls = list(self.requests_mock.calls)
            if calls:
                self._records[self.record_key] = calls

    @classmethod
    def tearDownClass(cls):
        if INTEGRATION_TESTING:
            with gzip.open(cls.recordpath, mode="wb") as file:
                pickle.dump(cls._records, file)

    @property
    def record_key(self):
        return self._testMethodName

    @staticmethod
    def _record_streaming_response_body(response):
        if not response.raw.closed:
            content = response.content
            response.raw = io.BytesIO(content)
        return response

    @classmethod
    def prepare_integration_test(cls):
        """This method can be used to check preconditions for integration tests"""
        pass


class DataServiceTestCase(ReplayResponsesTestCase):
    credentials = (
        CONSUMER_KEY or "pewh0CiBQ5Gl8BX7K2i8vww9tsr0",
        CONSUMER_SECRET or "cP58wJbenKRPaK9RKA8ODPxHmkw1",
    )

    @classmethod
    def tearDownClass(cls):
        if INTEGRATION_TESTING:
            key, secret = map(str.encode, cls.credentials)
            sensitive = b64encode(key + b":" + secret)
            xs = len(key) * b"x"
            crossed = b64encode(xs + b":" + xs)
            pickled = pickle.dumps(cls._records)
            sanitized = pickled.replace(sensitive, crossed)
            with gzip.open(cls.recordpath, mode="wb") as file:
                file.write(sanitized)


class FakeProduct:
    def __init__(self, id=None, collection=None):
        self._id = id or "product"
        self.collection = collection or FakeCollection()

    @property
    def md5(self):
        return None


class FakeCollection:
    def __init__(self, id=None):
        self._id = id or "collection"


class FakeCustomisation:
    def __init__(self, states_to_return, cid=None):
        self.states_to_return = iter(states_to_return)
        self.logfile = f"Fake Log: {states_to_return}"
        self._id = cid or "test"
        self.deleted = False

    @property
    def status(self):
        return next(self.states_to_return)

    @property
    def outputs(self):
        return ["prod1.nc", "prod2.nc"]

    def kill(self):
        pass

    def delete(self):
        if not self.deleted:
            self.deleted = True
        else:
            from eumdac.errors import CustomisationError

            raise CustomisationError("double deletion")

    @contextmanager
    def stream_output_iter_content(self, output, chunks=0):
        yield [bytes("test", "utf-8")]


class FakeTailor:
    def __init__(self):
        self.user_info = {
            "username": "test",
        }
        self.quota = {
            "data": {
                "test": {
                    "space_usage_percentage": 95,
                }
            }
        }

    def get_customisation(self, id):
        return FakeCustomisation(["QUEUED", "RUNNING", "DONE"], cid=id)

    def new_customisation(self, product, chain):
        return FakeCustomisation(["QUEUED", "RUNNING", "DONE"], cid=None)


class FakeStore:
    def get_product(self, col_id, p_id):
        return FakeProduct(id=p_id, collection=FakeCollection(id=col_id))



import io
import unittest
import re
import os
import datetime
import sys
from unittest.mock import patch, mock_open, MagicMock

import pytest

import eumdac
from eumdac.cli import cli
from eumdac.collection import Collection
from eumdac.errors import EumdacError
from eumdac.product import Product


def get_mock_collection(product_ids):
    mock_stream = io.BytesIO(b"ProductBytes")
    mock_stream.name = "MockProduct.zip"

    mock_collection = MagicMock(spec=Collection)
    mock_collection._id = "MockCollection"

    mock_collection_search_products = []
    for product_id in product_ids:
        mock_product = MagicMock(spec=Product)
        mock_product.open.return_value.__enter__.return_value = mock_stream
        mock_product._id = product_id
        mock_product.collection = mock_collection
        mock_collection_search_products.append(mock_product)

    mock_collection.search.return_value = mock_collection_search_products

    return mock_collection


class CliExit(Exception):
    pass


class CliError(Exception):
    pass


class TestCommandLineInterface(unittest.TestCase):
    credentials = ("abc", "xyz")

    def setUp(self):
        super().setUp()
        self.patch_argv()
        self.cli = cli
        self.get_mock("argparse.ArgumentParser.exit", new=MagicMock(side_effect=CliExit))
        self.mock_error = self.get_mock(
            "argparse.ArgumentParser.error", new=MagicMock(side_effect=CliError)
        )
        self.mock_stdin = self.get_mock("eumdac.cli.sys.stdin", new_callable=mock_open())
        self.mock_stdin.isatty.return_value = False

    @pytest.fixture(autouse=True)
    def propagate_pytest_fixtures(self, monkeypatch, tmp_path):
        self.monkeypatch = monkeypatch
        self.tmp_path = tmp_path

    def patch_argv(self):
        patcher = patch.object(sys, "argv", ["eumdac"])
        patcher.start()
        self.addCleanup(patcher.stop)

    def get_mock(self, target, **kwargs):
        patcher = patch(target, **kwargs)
        mock = patcher.start()
        self.addCleanup(patcher.stop)
        return mock

    @patch("eumdac.AccessToken")
    @patch("pathlib.Path.open", new_callable=mock_open)
    def test_set_credentials(self, mock_path_open, mock_AccessToken):
        consumer_key, consumer_secret = self.credentials
        command_line = ["--set-credentials", consumer_key, consumer_secret]
        with patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            with self.assertRaises(CliExit):
                self.cli(command_line)
            self.assertRegex(mock_stdout.getvalue(), r"Credentials are correct.")
        mock_path_open.return_value.__enter__().write.assert_called_once_with(
            f"{consumer_key},{consumer_secret}"
        )

    @patch("eumdac.cli.load_credentials")
    @patch("eumdac.cli.DataStore")
    def test_describe_list_collections(self, mock_DataStore, mock_load_credentials):
        mock_load_credentials.return_value = self.credentials
        collection = MagicMock(spec=Collection)
        collection.__str__.return_value = "MockCollection"
        collection.title = "MockTitle"
        mock_DataStore.return_value.collections = [collection]
        command_line = ["describe"]
        with patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            self.cli(command_line)
            self.assertRegex(mock_stdout.getvalue(), "MockCollection - MockTitle")

    def test_print_help(self):
        regex = re.compile(f"usage: (.*(\r)?\n)+EUMETSAT Data Access Client", re.MULTILINE)
        command_line = ["--help"]
        with patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            with self.assertRaises(CliExit):
                self.cli(command_line)
            self.assertRegex(mock_stdout.getvalue(), regex)

        # print help if no argument is given
        command_line = []
        with patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            self.cli(command_line)
            self.assertRegex(mock_stdout.getvalue(), regex)

    def test_print_version(self):
        command_line = ["--version"]
        with patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            with self.assertRaises(CliExit):
                self.cli(command_line)
            self.assertRegex(mock_stdout.getvalue(), re.escape(eumdac.__version__))

    def test_pipe_in(self):
        self.mock_stdin.read.return_value = "--version"
        with patch("eumdac.cli.sys.stdout", new_callable=io.StringIO) as mock_stdout:
            with self.assertRaises(CliExit):
                self.cli()
            self.assertRegex(mock_stdout.getvalue(), re.escape(eumdac.__version__))

    @patch("eumdac.cli.load_credentials")
    @patch("eumdac.cli.DataStore")
    def test_describe_collection(self, mock_DataStore, mock_load_credentials):
        mock_load_credentials.return_value = self.credentials
        collection = MagicMock(spec=Collection)
        collection.__str__.return_value = "MockCollection"
        collection.title = "MockTitle"
        collection.abstract = "MockAbstract"
        collection.metadata = {"properties": {"date": "2020/"}}
        mock_DataStore.return_value.get_collection.return_value = collection
        command_line = ["describe", "-c", "MockCollection"]
        with patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            self.cli(command_line)
            regex = re.compile(
                f"MockCollection.*MockTitle(.*(\r)?\n)+Date:(.*(\r)?\n)+MockAbstract",
                re.MULTILINE,
            )
            out = mock_stdout.getvalue()
            self.assertRegex(out, regex)

    @patch("eumdac.cli.load_credentials")
    def test_describe_product_without_collection(self, mock_load_credentials):
        mock_load_credentials.return_value = self.credentials
        command_line = ["describe", "-p", "MockProduct", "--debug"]
        with self.assertRaises(ValueError):
            self.cli(command_line)

    @patch("eumdac.cli.load_credentials")
    @patch("eumdac.cli.DataStore")
    def test_describe_product(self, mock_DataStore, mock_load_credentials):
        mock_load_credentials.return_value = self.credentials
        product = MagicMock(spec=Product)
        product.__str__.return_value = "MockProduct"
        product.collection = "MockCollection"
        product.satellite = "MockSatellite"
        product.instrument = "MockInstrument"
        product.acronym = "MockAcronym"
        product.sensing_start = datetime.datetime(2020, 1, 1)
        product.sensing_end = datetime.datetime(2020, 1, 2)
        product.size = 123456
        product.ingested = datetime.datetime(2020, 1, 2)
        product.md5 = "123456789"
        product.entries = [
            "product_file_1.nc",
            "product_file_2.nc",
            "product_file_3.nc",
            "product_group/entry1.nc",
            "product_group/entry2.nc",
            "product_group/entry3/file.nc",
            "metadata.xml",
        ]
        mock_DataStore.return_value.get_product.return_value = product
        expected_lines = [
            "MockCollection - MockProduct",
            "Platform: MockSatellite",
            "Instrument: MockInstrument",
            "Acronym: MockAcronym",
            "Orbit: LEO",
            "Sensing Start: 2020-01-01T00:00:00.000Z",
            "Sensing End: 2020-01-02T00:00:00.000Z",
            "Size: 123456 KB",
            "Published: 2020-01-02T00:00:00.000Z",
            "MD5: 123456789",
            "SIP Entries:",
            "product_file_1.nc",
            "product_file_2.nc",
            "product_file_3.nc",
            "product_group/",
            "- entry1.nc",
            "- entry2.nc",
            "- entry3/file.nc",
            "metadata.xml",
            "",
        ]
        command_line = ["describe", "-c", "MockCollection", "-p", "MockProduct"]
        with patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            self.cli(command_line)
            for line in expected_lines:
                self.assertIn(line, mock_stdout.getvalue())

    @patch("eumdac.cli.load_credentials")
    @patch("eumdac.cli.DataStore")
    def test_describe_product_flat(self, mock_DataStore, mock_load_credentials):
        mock_load_credentials.return_value = self.credentials
        product = MagicMock(spec=Product)
        product.__str__.return_value = "MockProduct"
        product.collection = "MockCollection"
        product.satellite = "MockSatellite"
        product.instrument = "MockInstrument"
        product.acronym = "MockAcronym"
        product.sensing_start = datetime.datetime(2020, 1, 1)
        product.sensing_end = datetime.datetime(2020, 1, 2)
        product.size = 123456
        product.ingested = datetime.datetime(2020, 1, 2)
        product.md5 = "123456789"
        product.entries = [
            "product_file_1.nc",
            "product_file_2.nc",
            "product_file_3.nc",
            "product_group/entry1.nc",
            "product_group/entry2.nc",
            "product_group/entry3/file.nc",
            "metadata.xml",
        ]
        mock_DataStore.return_value.get_product.return_value = product
        expected_lines = [
            "MockCollection - MockProduct",
            "Platform: MockSatellite",
            "Instrument: MockInstrument",
            "Acronym: MockAcronym",
            "Orbit: LEO",
            "Sensing Start: 2020-01-01T00:00:00.000Z",
            "Sensing End: 2020-01-02T00:00:00.000Z",
            "Size: 123456 KB",
            "Published: 2020-01-02T00:00:00.000Z",
            "MD5: 123456789",
            "SIP Entries:",
            "product_file_1.nc",
            "product_file_2.nc",
            "product_file_3.nc",
            "product_group/entry1.nc",
            "product_group/entry2.nc",
            "product_group/entry3/file.nc",
            "metadata.xml",
            "",
        ]
        command_line = ["describe", "-c", "MockCollection", "-p", "MockProduct", "--flat"]
        with patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            self.cli(command_line)
            for line in expected_lines:
                self.assertIn(line, mock_stdout.getvalue())

    @patch("eumdac.cli.AccessToken")
    @patch("eumdac.cli.DataStore")
    @patch("eumdac.cli.load_credentials")
    def test_load_credentials(self, mock_load_credentials, mock_DataStore, mock_AccessToken):
        mock_load_credentials.return_value = ("CONSUMER_KEY", "CONSUMER_SECRET")
        mock_DataStore.return_value.collections = []
        command_line = ["describe"]
        self.cli(command_line)
        mock_AccessToken.assert_called_once_with(("CONSUMER_KEY", "CONSUMER_SECRET"))

    @patch("pathlib.Path.read_text")
    @patch("eumdac.cli.DataStore")
    def test_load_credentials_file_not_found(self, mock_DataStore, mock_read_text):
        mock_read_text.side_effect = FileNotFoundError
        command_line = (
            "download -c EO:EUM:DAT:MSG:MSG15-RSS "
            "--time-range 2021-09-20 2022-09-30 --limit 1 --debug"
        ).split()
        with self.assertRaises(EumdacError):
            self.cli(command_line)

    @patch("eumdac.cli.get_credentials_path")
    @patch("eumdac.cli.AccessToken")
    @patch("eumdac.cli.DataStore")
    def test_load_credentials_corrupted_file(
        self, mock_DataStore, mock_AccessToken, mock_get_credentials_path
    ):
        mock_get_credentials_path.return_value.read_text.return_value = "corrupted!"
        mock_DataStore.return_value.collections = []
        command_line = ["describe", "--debug"]
        with self.assertRaises(EumdacError):
            self.cli(command_line)

    @patch("eumdac.cli.get_datastore")
    @patch("eumdac.cli.load_credentials")
    def test_search(self, _, mock_get_datastore):
        mock_collection = MagicMock(spec=Collection)
        mock_collection.search.return_value = [
            "MSG4-SEVI-MSGCLMK-0100-0100-20200301121500.000000000Z-NA",
            "MSG4-SEVI-MSGCLMK-0100-0100-20200301120000.000000000Z-NA",
            "MSG4-SEVI-MSGCLMK-0100-0100-20200301114500.000000000Z-NA",
            "MSG4-SEVI-MSGCLMK-0100-0100-20200301113000.000000000Z-NA",
            "MSG4-SEVI-MSGCLMK-0100-0100-20200301111500.000000000Z-NA",
            "MSG4-SEVI-MSGCLMK-0100-0100-20200301110000.000000000Z-NA",
            "MSG4-SEVI-MSGCLMK-0100-0100-20200301104500.000000000Z-NA",
            "MSG4-SEVI-MSGCLMK-0100-0100-20200301103000.000000000Z-NA",
            "MSG4-SEVI-MSGCLMK-0100-0100-20200301101500.000000000Z-NA",
            "MSG4-SEVI-MSGCLMK-0100-0100-20200301100000.000000000Z-NA",
        ]
        mock_DataStore = MagicMock(spec=eumdac.DataStore)
        mock_DataStore.get_collection.return_value = mock_collection
        mock_get_datastore.return_value = mock_DataStore
        command_line = [
            "search",
            "-s",
            "2020-03-01",
            "-e",
            "2020-03-01T12:15",
            "-c",
            "EO:EUM:DAT:MSG:CLM",
            "--limit",
            "3",
        ]
        expected_lines = [
            "MSG4-SEVI-MSGCLMK-0100-0100-20200301121500.000000000Z-NA",
            "MSG4-SEVI-MSGCLMK-0100-0100-20200301120000.000000000Z-NA",
            "MSG4-SEVI-MSGCLMK-0100-0100-20200301114500.000000000Z-NA",
            "",
        ]
        expected_output = "\n".join(expected_lines)
        with patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            self.cli(command_line)
            self.assertEqual(mock_stdout.getvalue(), expected_output)

    @patch("eumdac.cli.get_datastore")
    @patch("eumdac.cli.load_credentials")
    def test_search_bbox(self, _, mock_get_datastore):
        mock_collection = MagicMock(spec=Collection)
        mock_collection.search.return_value = [
            "MockProduct-a",
            "MockProduct-b",
            "MockProduct-c",
            "MockProduct-d",
        ]
        mock_DataStore = MagicMock(spec=eumdac.DataStore)
        mock_DataStore.get_collection.return_value = mock_collection
        mock_get_datastore.return_value = mock_DataStore

        command_line = [
            "search",
            "-c",
            "MockCollection",
            "--bbox",
            "2.0",
            "10.0",
            "10.0",
            "52.0",
            "--limit",
            "3",
        ]
        self.cli(command_line)
        print(mock_collection.call_count)
        mock_collection.search.assert_called_with(bbox="2.0,10.0,10.0,52.0", set="brief")

    @patch("eumdac.cli.load_credentials")
    def test_local_tailor_set(self, mock_load_credentials):
        mock_load_credentials.return_value = self.credentials
        tailor_url = "http://NON_EXISTANT:40000"
        tailor_id = "local"

        def mock_get_url_path():
            return self.tmp_path

        self.monkeypatch.setattr("eumdac.local_tailor.get_url_path", mock_get_url_path)
        command_line = ["local-tailor", "set", tailor_id, tailor_url]
        with patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            self.cli(command_line)

        command_line = ["local-tailor", "remove", tailor_id]
        with patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            self.cli(command_line)

    @patch("eumdac.cli.load_credentials")
    def test_local_tailor_instances(self, mock_load_credentials):
        mock_load_credentials.return_value = self.credentials
        tailor_url = "http://NON_EXISTANT:40000"
        tailor_id = "local"

        def mock_get_url_path():
            return self.tmp_path

        self.monkeypatch.setattr("eumdac.local_tailor.get_url_path", mock_get_url_path)

        command_line = ["local-tailor", "set", tailor_id, tailor_url]
        with patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            self.cli(command_line)

        command_line = ["local-tailor", "instances"]
        with patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            self.cli(command_line)
            self.assertRegex(mock_stdout.getvalue(), f".*{tailor_url}.*")
            self.assertRegex(mock_stdout.getvalue(), f".*{tailor_id}.*")
            self.assertRegex(mock_stdout.getvalue(), f".*OFFLINE.*")

        command_line = ["local-tailor", "remove", tailor_id]
        with patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            self.cli(command_line)

    @patch("eumdac.cli.load_credentials")
    def test_local_tailor_show(self, mock_load_credentials):
        mock_load_credentials.return_value = self.credentials
        tailor_url = "http://NON_EXISTANT:40000"
        tailor_id = "local"

        def mock_get_url_path():
            return self.tmp_path

        self.monkeypatch.setattr("eumdac.local_tailor.get_url_path", mock_get_url_path)

        command_line = ["local-tailor", "set", tailor_id, tailor_url]
        with patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            self.cli(command_line)

        command_line = ["local-tailor", "show", tailor_id]
        with patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            self.cli(command_line)
            self.assertRegex(mock_stdout.getvalue(), f".*{tailor_url}.*")
            self.assertRegex(mock_stdout.getvalue(), f".*{tailor_id}.*")
            self.assertRegex(mock_stdout.getvalue(), f".*OFFLINE.*")

        command_line = ["local-tailor", "remove", tailor_id]
        with patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            self.cli(command_line)
